import itertools

import numpy as np
import collections

from .. import Numputils as nput

__all__ = [
    "RedundantCoordinateGenerator"
]


class RedundantCoordinateGenerator:

    def __init__(self,
                 coordinate_specs, angle_ordering='ijk',
                 untransformed_coordinates=None, masses=None,
                 relocalize=False,
                 **opts):
        self.specs = coordinate_specs
        self.untransformed_coordinates = untransformed_coordinates
        self.relocalize = relocalize
        self.masses = masses
        self.opts = dict(
            opts,
            angle_ordering=angle_ordering
        )

    @classmethod
    def _pad_redund_tf(cls, redund_tf, n):
        ndim = redund_tf.ndim - 2
        eye = nput.identity_tensors(redund_tf.shape[:-2], n)
        return np.concatenate(
            [
                np.pad(eye, ([[0, 0]] * ndim) + [[0, redund_tf.shape[-2]], [0, 0]]),
                np.pad(redund_tf, ([[0, 0]] * ndim) + [[n, 0], [0, 0]])
            ],
            axis=-1
        )

    @classmethod
    def _relocalize_tf(cls, redund_tf, untransformed_coordinates=None):
        n = redund_tf.shape[-1]
        if untransformed_coordinates is None:
            # target = np.pad(np.eye(n), [[0, 173 - 108], [0, 0]])
            ndim = redund_tf.ndim - 2
            eye = nput.identity_tensors(redund_tf.shape[:-2], n)
            target = np.pad(eye, ([[0, 0]] * ndim) + [[0, redund_tf.shape[-2] - n], [0, 0]])
        else:
            untransformed_coordinates = np.asanyarray(untransformed_coordinates)
            coords = np.concatenate([
                untransformed_coordinates,
                np.delete(np.arange(redund_tf.shape[-2]), untransformed_coordinates)
                ])[:n]
            target = np.moveaxis(np.zeros(redund_tf.shape), -1, -2)
            idx = nput.vector_ix(target.shape[:-1], coords[:, np.newaxis])
            target[idx] = 1
            target = np.moveaxis(target, -1, -2)
        loc = np.linalg.lstsq(redund_tf, target, rcond=None)
        U, s, V = np.linalg.svd(loc[0])
        R = U @ V
        return redund_tf @ R


    @classmethod
    def base_redundant_transformation(cls, expansion,
                                      untransformed_coordinates=None,
                                      masses=None,
                                      relocalize=False
                                      ):
        conv = np.asanyarray(expansion[0])
        if masses is not None:
            masses = np.asanyarray(masses)
            if len(masses) == conv.shape[-2] // 3:
                masses = np.repeat(masses, 3)
            masses = np.diag(1 / np.sqrt(masses))
            if conv.ndim > 2:
                masses = np.broadcast_to(
                    np.expand_dims(masses, list(range(conv.ndim - 2))),
                    conv.shape[:-2] + masses.shape
                )
            conv = masses @ conv

        if untransformed_coordinates is not None:
            transformed_coords = np.setdiff1d(np.arange(conv.shape[-1]), untransformed_coordinates)
            ut_conv = conv[..., untransformed_coordinates]
            conv = conv[..., transformed_coords]

            # project out contributions along untransformed coordinates to ensure
            # dimension of space remains unchanged
            ut_conv_norm = nput.vec_normalize(np.moveaxis(ut_conv, -1, -2))
            proj = np.moveaxis(ut_conv_norm, -1, -2) @ ut_conv_norm
            proj = nput.identity_tensors(proj.shape[:-2], proj.shape[-1]) - proj

            conv = np.concatenate([ut_conv, proj @ conv], axis=-1)

        G_internal = np.moveaxis(conv, -1, -2) @ conv
        redund_vals, redund_tf = np.linalg.eigh(G_internal)
        redund_pos = np.where(np.abs(redund_vals) > 1e-10)
        if redund_vals.ndim > 1:
            redund_tf = nput.take_where_groups(redund_tf, redund_pos)
        else:
            redund_tf = redund_tf[:, redund_pos[0]]
        if isinstance(redund_tf, np.ndarray):
            redund_tf = np.flip(redund_tf, axis=-1)
        else:
            redund_tf = [
                np.flip(tf, axis=-1)
                for tf in redund_tf
            ]

        # if transformed_coords is not None:
        #     n = len(untransformed_coordinates)
        #     if isinstance(redund_tf, np.ndarray):
        #         redund_tf = cls._pad_redund_tf(redund_tf, n)
        #     else:
        #         redund_tf = [
        #             cls._pad_redund_tf(tf, n)
        #             for tf in redund_tf
        #         ]

        if relocalize:
            if untransformed_coordinates is not None:
                perm = np.concatenate([untransformed_coordinates,
                                       np.delete(np.arange(redund_tf.shape[-2]), untransformed_coordinates)
                                       ])
                perm = np.argsort(perm)
                if isinstance(redund_tf, np.ndarray):
                    redund_tf = redund_tf[..., perm, :]
                else:
                    redund_tf = [
                        redund_tf[..., perm, :]
                        for tf in redund_tf
                    ]
            # provide some facility for rearranging coords?
            if isinstance(redund_tf, np.ndarray):
                redund_tf = cls._relocalize_tf(redund_tf, untransformed_coordinates=untransformed_coordinates)
            else:
                redund_tf = [
                    cls._relocalize_tf(tf, untransformed_coordinates=untransformed_coordinates)
                    for tf in redund_tf
                ]



        return redund_tf


    @classmethod
    def get_redundant_transformation(cls, base_expansions, untransformed_coordinates=None, masses=None,
                                     relocalize=False):
        if isinstance(base_expansions, np.ndarray) and base_expansions.ndim == 2:
            base_expansions = [base_expansions]
            base_inv = None
        elif (
                len(base_expansions) == 2
                and not nput.is_numeric_array_like(base_expansions[0], ndim=2)
        ):
            base_expansions, base_inv = base_expansions
        else:
            base_inv = None
        redund_tf = cls.base_redundant_transformation(base_expansions,
                                                      untransformed_coordinates=untransformed_coordinates,
                                                      masses=masses,
                                                      relocalize=relocalize
                                                      )

        # dQ/dR, which we can transform with dR/dX to get dQ/dX
        if isinstance(redund_tf, np.ndarray):
            redund_expansions = nput.tensor_reexpand(base_expansions, [redund_tf], order=len(base_expansions))
            if base_inv is not None:
                redund_inv = nput.tensor_reexpand([redund_tf.T], base_inv, order=len(base_inv))
                redund_expansions = (redund_expansions, redund_inv)
        else:
            redund_expansions = [
                nput.tensor_reexpand(base_expansions, [tf], order=len(base_expansions))
                for tf in redund_tf
            ]
            if base_inv is not None:
                redund_inv = [
                    nput.tensor_reexpand([tf.T], base_inv, order=len(base_inv))
                    for tf in redund_tf
                ]
                redund_expansions = (redund_expansions, redund_inv)

        return redund_tf, redund_expansions

    def compute_redundant_expansions(self,
                                     coords,
                                     order=None,
                                     untransformed_coordinates=None,
                                     expansions=None,
                                     relocalize=None):
        coords = np.asanyarray(coords)
        if order is None:
            opts = dict(dict(order=1), **self.opts)
        else:
            opts = dict(self.opts, order=order)
        if untransformed_coordinates is None:
            untransformed_coordinates = self.untransformed_coordinates
        if relocalize is None:
            relocalize = self.relocalize
        if expansions is None:
            base_expansions = nput.internal_coordinate_tensors(coords, self.specs, **opts)[1:]
        else:
            base_expansions = expansions
        return self.get_redundant_transformation(base_expansions,
                                                 untransformed_coordinates=untransformed_coordinates,
                                                 masses=self.masses,
                                                 relocalize=relocalize
                                                 )

    @classmethod
    def _prune_coords_svd(cls, b_mat, svd_cutoff=5e-2, sort=True, fixed_vecs=None):
        # turns out, equivalent to finding maximimum loc in eigenvectors of G
        if fixed_vecs is not None:
            transformed_coords = np.delete(np.arange(b_mat.shape[-1]), fixed_vecs)
            ut_conv = b_mat[..., fixed_vecs]
            conv = b_mat[..., transformed_coords]

            # project out contributions along untransformed coordinates to ensure
            # dimension of space remains unchanged
            ut_conv = nput.vec_normalize(ut_conv)
            proj = ut_conv @ np.moveaxis(ut_conv, -1, -2)
            proj = nput.identity_tensors(proj.shape[:-2], proj.shape[-1]) - proj

            b_mat = proj @ conv

        _, s, Q = np.linalg.svd(b_mat)
        pos = np.where(s > 1e-8)
        loc_val = np.max(Q[pos]**2, axis=0)
        coords = np.where(loc_val > svd_cutoff)[0]
        if sort:
            coords = coords[np.argsort(-loc_val[coords,],)]
        if fixed_vecs is not None:
            coords = transformed_coords[coords,]
            coords = np.concatenate([fixed_vecs, coords])
        return coords

    @classmethod
    def _prune_coords_loc(cls, b_mat, loc_cutoff=.33, sort=True, fixed_vecs=None):
        # if fixed_vecs is not None:
        #     transformed_coords = np.delete(np.arange(b_mat.shape[-1]), fixed_vecs)
        #     ut_conv = b_mat[..., fixed_vecs]
        #     b_mat = b_mat[..., transformed_coords]
        #
        #     # # project out contributions along untransformed coordinates to ensure
        #     # # dimension of space remains unchanged
        #     # ut_conv = nput.vec_normalize(ut_conv)
        #     # proj = ut_conv @ np.moveaxis(ut_conv, -1, -2)
        #     # proj = nput.identity_tensors(proj.shape[:-2], proj.shape[-1]) - proj
        #     #
        #     # b_mat = proj @ conv

        s, Q = np.linalg.eigh(b_mat.T @ b_mat)
        pos = np.where(s > 1e-8)[0]
        Q = Q[:, pos]
        loc_mat = np.linalg.lstsq(Q, np.eye(Q.shape[-2]), rcond=None)
        loc_val = np.diag(Q @ loc_mat[0])
        coords = np.where(loc_val > loc_cutoff)[0]
        if fixed_vecs is not None:
            coords = np.setdiff1d(coords, fixed_vecs)
        if sort:
            coords = coords[np.argsort(-loc_val[coords,], )]
        if fixed_vecs is not None:
            coords = np.concatenate([fixed_vecs, coords])

            # have to run this back again in case these added coords break the localization
            # a third run back isn't work it since that will just repeat what's found here
            b_mat = b_mat[:, coords]
            s, Q = np.linalg.eigh(b_mat.T @ b_mat)
            pos = np.where(s > 1e-8)[0]
            Q = Q[:, pos]
            loc_mat = np.linalg.lstsq(Q, np.eye(Q.shape[-2]), rcond=None)
            loc_val = np.diag(Q @ loc_mat[0])
            subcoords = np.setdiff1d(np.where(loc_val > loc_cutoff)[0], np.arange(len(fixed_vecs)))
            if sort:
                subcoords = subcoords[np.argsort(-loc_val[subcoords,], )]
            coords = np.concatenate([coords[:len(fixed_vecs)], coords[subcoords]])

        return coords

    @classmethod
    def _prune_coords_gs(cls, b_mat, fixed_vecs=None, core_scaling=1e3, max_condition_number=1e8):
        if fixed_vecs is None:
            _, s, Q = np.linalg.svd(b_mat)
            pos = np.where(s > 0)
            fixed_vecs = np.where(np.linalg.norm(Q[pos], axis=0) > .9)[0]
        fixed_vecs = np.array(fixed_vecs)
        core_mask = np.full(b_mat.shape[1], False)
        core_mask[fixed_vecs] = True
        all_mask = core_mask.copy()
        rem_pos = np.setdiff1d(np.arange(b_mat.shape[1]), fixed_vecs)
        base_evals = np.linalg.eigvalsh(b_mat[:, core_mask].T @ b_mat[:, core_mask])
        base_cond = base_evals[-1] / base_evals[0] * core_scaling
        for r in rem_pos:
            core_mask[r] = True
            evals = np.linalg.eigvalsh(b_mat[:, core_mask].T @ b_mat[:, core_mask])
            new_cond = abs(evals[-1] / evals[0])
            if new_cond < max_condition_number:
                all_mask[r] = True
                if new_cond > base_cond:
                    core_mask[r] = False
            else:
                core_mask[r] = False

        return np.where(all_mask)[0]

    @classmethod
    def prune_coordinate_specs(cls, expansion,
                               masses=None,
                               untransformed_coordinates=None,
                               pruning_mode='loc',
                               **opts
                               ):
        conv = np.asanyarray(expansion[0])
        masses = np.asanyarray(masses)
        if len(masses) == conv.shape[-2] // 3:
            masses = np.repeat(masses, 3)
        masses = np.diag(1 / np.sqrt(masses))
        b_mat = masses @ conv

        if pruning_mode == 'svd':
            return cls._prune_coords_svd(b_mat, fixed_vecs=untransformed_coordinates, **opts)
        elif pruning_mode == 'loc':
            return cls._prune_coords_loc(b_mat, fixed_vecs=untransformed_coordinates, **opts)
        elif pruning_mode == 'gs':
            return cls._prune_coords_gs(b_mat, fixed_vecs=untransformed_coordinates, **opts)
        else:
            raise ValueError(f"don't understand pruning mode {pruning_mode}")

class MultiOriginCoordinates:

    def __init__(self, origins, zmats):
        ...




import sys, os, pathlib, abc, enum, types, string, inspect
import ast, functools, re, fnmatch, collections
import typing, itertools, builtins
from typing import *

from ... import Devutils as dev
from .ObjectWalker import *
__reload_hook__ = [".ObjectWalker"]

__all__ = [
    "TemplateFormatter",
    "FormatDirective",
    "TemplateFormatDirective",
    "TemplateOps",
    "TemplateEngine",
    "ResourceLocator",
    "TemplateResourceExtractor",
    "TemplateWalker",
    "TemplateHandler",
    "ModuleTemplateHandler",
    "ClassTemplateHandler",
    "FunctionTemplateHandler",
    "MethodTemplateHandler",
    "ObjectTemplateHandler",
    "IndexTemplateHandler",
    "TemplateInterfaceEngine",
    "TemplateInterfaceFormatter"
]

class TemplateOps:
    @staticmethod
    def loop(caller: typing.Callable, *args, joiner="", formatter=None, **kwargs):
        if len(kwargs) == 0:
            res = [caller(*a) for a in zip(*args)]
        elif len(args) == 0:
            res = [
                caller(**{k:v for k,v in zip(kwargs.keys(), kv)})
                for kv in zip(*kwargs.values())
            ]
        else:
            res = [
                caller(*a, **{k:v for k,v in zip(kwargs.keys(), kv)})
                for a,kv in zip(zip(*args), zip(*kwargs.values()))
            ]
        if joiner is not None:
            res = joiner.join(res)
        return res
    @classmethod
    def loop_template(cls, template:str, *args, joiner="", formatter=None, **kwargs):
        return cls.loop(
            template.format,
            *args,
            joiner=joiner,
            formatter=formatter,
            **kwargs
        )
    @staticmethod
    def join(*args, joiner=" ", formatter=None):
        if len(args) == 1 and not isinstance(args[0], str):
            args = args[0]
        return joiner.join(args)
    @classmethod
    def load(cls, template, formatter=None):
        return formatter.load_template(template)
    @classmethod
    def include(cls, template, formatter=None):
        return formatter.vformat(formatter.load_template(template), (), formatter.format_parameters)
    @classmethod
    def apply(cls, template, *args, formatter=None, **kwargs):
        if formatter is None:
            raise ValueError("`{}` can't be `None`".format('formatter'))
        return formatter.format(template, *args, **kwargs)
    @classmethod
    def nonempty(cls, data, formatter=None):
        return data is not None and len(data) > 0
    @classmethod
    def wrap(cls, fn):
        @functools.wraps(fn)
        def f(*args, formatter=None, **kwargs):
            return fn(*args, **kwargs)
        return f
    @staticmethod
    def cleandoc(txt, formatter=None):
        return inspect.cleandoc(txt)
    @staticmethod
    def wrap_str(obj, formatter=None):
        txt = str(obj)
        txt = txt.replace('"', '\\"').replace("'", "\\'")
        if '\n' in txt:
            txt = '"""'+txt+'"""'
        return txt
    @staticmethod
    def optional(key, default="", formatter=None):
        return formatter.format_parameters.get(key, default)

class FormatDirective(enum.Enum):
    """
    Base class for directives -- shouldn't be an enum really...
    """
    def __init__(self, name, callback=None):
        self.key = name
        if isinstance(callback, str):
            callback = callback.format
        elif isinstance(callback, (staticmethod, classmethod, property)):
            callback = callback.__get__(self)
        self.callback = callback if not isinstance(callback, str) else callback.format
    def _call(self, *data, **kwargs):
        return self.callback(*data, **kwargs)
    @classmethod
    def _keymap(cls):
        if not hasattr(cls, '_keymap_dict'):
            cls._keymap_dict = None
        return cls._keymap_dict
    @classmethod
    def _load(cls, name:str):
        k = cls._keymap()
        if k is None:
            cls._keymap_dict = {c.key:c for c in cls}
            k = cls._keymap_dict
        return k[name]
    @classmethod
    def extend(cls, *others):
        vals = {
            o.name:o.value
            for c in (cls,) + others
            for o in c
        }
        return FormatDirective(cls.__name__, vals)

class TemplateFormatDirective(FormatDirective):
    Loop = "loop", TemplateOps.loop
    LoopTemplate = "loop_template", TemplateOps.loop_template
    Join = "join", TemplateOps.join
    Load = "load", TemplateOps.load
    Include = "include", TemplateOps.include
    Apply = "apply", TemplateOps.apply
    NonEmpty = "nonempty", TemplateOps.nonempty
    CleanDoc = "cleandoc", TemplateOps.cleandoc
    Optional = "optional", TemplateOps.optional

    Str = "str", TemplateOps.wrap_str
    Int = "int", TemplateOps.wrap(int)
    Float = "float", TemplateOps.wrap(float)
    Round = "round", TemplateOps.wrap(round)
    Len = "len", TemplateOps.wrap(len)
    Dict = "dict", TemplateOps.wrap(dict)
    List = "list", TemplateOps.wrap(list)
    Tuple = "tuple", TemplateOps.wrap(tuple)
    Set = "set", TemplateOps.wrap(set)

class TemplateFormatterError(ValueError):
    ...
class TemplateASTEvaluator:
    def __init__(self, formatter, directives, format_parameters:dict):
        self.formatter = formatter
        self.directives = directives
        self.format_parameters = format_parameters
    def handle_comprehension(self, g, expr, callback):
        target = g.target.id
        restore = False
        if target in self.format_parameters:
            restore = True
            old = self.format_parameters[target]
        try:
            itt = self.evaluate_node(g.iter)
            for v in itt:
                self.format_parameters[target] = v
                if all(self.evaluate_node(e) for e in g.ifs):
                    callback(self.evaluate_node(expr))
        finally:
            if restore:
                self.format_parameters[target] = old
            else:
                if target in self.format_parameters:
                    del self.format_parameters[target]
    def evaluate_node(self, node:typing.Union[ast.AST,ast.expr,tuple]):
        if isinstance(node, tuple):
            return tuple(self.evaluate_node(n) for n in node)
        elif isinstance(node, ast.Module):
            bits = []
            for e in node.body:
                res = self.evaluate_node(e)
                if res is None:
                    res = ""
                bits.append(str(res))
            return "".join(bits)
        elif isinstance(node, ast.Name) and isinstance(node.ctx, ast.Load):
            name = node.id
            try:
                val = self.format_parameters[name]
            except KeyError:
                need_raise = False
                try:
                    val = getattr(builtins, name)
                except AttributeError:
                    need_raise = True
                if need_raise:
                    raise
            return val
        elif isinstance(node, ast.Attribute) and isinstance(node.ctx, ast.Load):
            return getattr(self.evaluate_node(node.value), node.attr)
        elif isinstance(node, ast.Assign):
            if len(node.targets) > 1:
                raise TemplateFormatterError("Template assignments are restricted to single reassignments")
            if not isinstance(node.targets[0], ast.Name):
                raise TemplateFormatterError("Template assignments are restricted to variable names")
            self.format_parameters[node.targets[0].id] = self.evaluate_node(node.value)
        elif isinstance(node, ast.Expr):
            return self.evaluate_node(node.value)
        elif isinstance(node, ast.Constant):
            return node.value
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.List):
            return [self.evaluate_node(e) for e in node.elts]
        elif isinstance(node, ast.ListComp):
            l = []
            for g in node.generators:
                if isinstance(g, ast.comprehension):
                    self.handle_comprehension(
                        g,
                        node.elt,
                        l.append
                    )
            return l
        elif isinstance(node, ast.Tuple):
            return tuple(self.evaluate_node(e) for e in node.elts)
        elif isinstance(node, ast.Set):
            return {self.evaluate_node(e) for e in node.elts}
        elif isinstance(node, ast.SetComp):
            l = set
            for g in node.generators:
                if isinstance(g, ast.comprehension):
                    self.handle_comprehension(
                        g,
                        node.elt,
                        l.add
                    )
            return l
        elif isinstance(node, ast.Dict):
            return {self.evaluate_node(k):self.evaluate_node(v) for k,v in zip(node.keys, node.values)}
        elif isinstance(node, ast.DictComp):
            l = {}
            add = lambda kv:l.__setitem__(*kv)
            for g in node.generators:
                if isinstance(g, ast.comprehension):
                    self.handle_comprehension(
                        g,
                        (node.key, node.value),
                        add
                    )
            return l

        elif isinstance(node, ast.UnaryOp):
            val = self.evaluate_node(node.operand)
            if isinstance(node.op, ast.Not):
                return not val
            else:
                raise TemplateFormatterError("unsupported operation {}".format(ast.dump(node.op)))
        elif isinstance(node, ast.BinOp):
            left = self.evaluate_node(node.left)
            right = self.evaluate_node(node.right)
            if isinstance(node.op, ast.Add):
                return left + right
            elif isinstance(node.op, ast.Sub):
                return left - right
            elif isinstance(node.op, ast.Mult):
                return left * right
            elif isinstance(node.op, ast.Div):
                return left / right
            elif isinstance(node.op, ast.FloorDiv):
                return left // right
            elif isinstance(node.op, ast.MatMult):
                return left @ right
            elif isinstance(node.op, ast.And):
                return left and right
            elif isinstance(node.op, ast.Or):
                return left or right
            elif isinstance(node.op, ast.BitOr):
                return left | right
            elif isinstance(node.op, ast.BitAnd):
                return left & right
            elif isinstance(node.op, ast.Pow):
                return left ** right
            else:
                raise TemplateFormatterError("unsupported operation {}".format(ast.dump(node.op)))
        elif isinstance(node, ast.BoolOp):
            if isinstance(node.op, ast.Or):
                return self.evaluate_node(node.values[0]) or self.evaluate_node(node.values[1])
            elif isinstance(node.op, ast.And):
                return self.evaluate_node(node.values[0]) and self.evaluate_node(node.values[1])
            else:
                raise TemplateFormatterError("unsupported operation {}".format(ast.dump(node.op)))
        elif isinstance(node, ast.Compare):
            left = self.evaluate_node(node.left)
            op = node.ops[0]
            right = self.evaluate_node(node.comparators[0])
            if isinstance(op, ast.Eq):
                return left == right
            elif isinstance(op, ast.NotEq):
                return left != right
            elif isinstance(op, ast.Lt):
                return left < right
            elif isinstance(op, ast.Gt):
                return left > right
            elif isinstance(op, ast.LtE):
                return left <= right
            elif isinstance(op, ast.GtE):
                return left >= right
            elif isinstance(op, ast.In):
                return left in right
            elif isinstance(op, ast.NotIn):
                return left not in right
            elif isinstance(op, ast.Is):
                return left is right
            elif isinstance(op, ast.IsNot):
                return left is not right
            else:
                raise TemplateFormatterError("unsupported comparison {}".format(ast.dump(op)))
        elif isinstance(node, ast.IfExp):
            if self.evaluate_node(node.test):
                return self.evaluate_node(node.body)
            else:
                return self.evaluate_node(node.orelse)
        elif isinstance(node, ast.Subscript):
            return self.evaluate_node(node.value).__getitem__(self.evaluate_node(node.slice))
        elif isinstance(node, ast.Index):
            return self.evaluate_node(node.value)
        elif isinstance(node, ast.Call):
            if isinstance(node.func, ast.Attribute): # subattr of some object...I guess we can call it?
                directive = self.evaluate_node(node.func)
            directive = None
            if isinstance(node.func, ast.Name):
                name = node.func.id
                if name == 'raw':
                    return TemplateOps.wrap_str(node.args[0])
                elif name == 'assign':
                    self.format_parameters[self.evaluate_node(node.args[0])] = self.evaluate_node(node.args[1])
                    return ""
                try:
                    directive = self.directives._load(name)
                except KeyError:
                    need_raise = name in {'open'}
                    if not need_raise:
                        try:
                            directive = getattr(builtins, name)
                        except AttributeError:
                            need_raise = True
                    if need_raise:
                        raise
                else:
                    directive = lambda *a, _d=directive, **k: _d._call(*a, formatter=self.formatter, **k)
            if directive is None:
                directive = self.evaluate_node(node.func)
            args = [self.evaluate_node(a) for a in node.args]
            kwargs = {k.arg: self.evaluate_node(k.value) for k in node.keywords}
            return directive(*args, **kwargs)
        else:
            raise TemplateFormatterError("Node {} unsupported".format(ast.dump(node)))
class TemplateFormatter(string.Formatter):
    """
    Provides a formatter for fields that allows for
    the inclusion of standard Bootstrap HTML elements
    alongside the classic formatting
    """
    max_recusion=6
    directives = TemplateFormatDirective
    class frozendict(dict):
        def __setitem__(self, key, value):
            raise TypeError("`frozendict` is immutable")
    def __init__(self, templates):
        self.__templates = self.frozendict(templates)
        self._fmt_stack = []
    @property
    def format_parameters(self):
        return self._fmt_stack[-1] if len(self._fmt_stack) > 0 else None
    @property
    def templates(self):
        return self.__templates
    @property
    def special_callbacks(self):
        return {"%":self.apply_eval_tree, "$":self.apply_directive_tree, "#":self.apply_comment, 'raw$':self.apply_raw, 'assign%':self.apply_assignment}
    @property
    def callback_map(self):
        return dict(
            self.special_callbacks,
            **{d.key+"$":self.apply_directive for d in self.directives}
        )

    def apply_eval_tree(self, _, spec) -> str:
        tree = ast.parse(inspect.cleandoc(spec))
        ev = TemplateASTEvaluator(self, self.directives, self.format_parameters).evaluate_node(tree)
        if ev is None:
            ev = ""
        return ev
    def apply_directive_tree(self, _, spec) -> str:
        return self.apply_eval_tree(_, "("+spec+")")
    def apply_assignment(self, key, spec) -> str:
        key, val = spec.split("=", 1)
        self.format_parameters[key] = val
        return ""
    def apply_raw(self, key, spec) -> str:
        return spec
    def apply_comment(self, key, spec) -> str:
        return ""
    def apply_directive(self, key, spec) -> str:
        return self.apply_directive_tree(
            key,
            "{}({})".format(key.strip("$"), spec)
        )
    def format_field(self, value: Any, format_spec: str) -> str:
        if self.format_parameters is None:
            raise NotImplementedError("{}.{} called outside of `vformat`".format(
                type(self).__name__,
                'format_field'
            ))
        callback = (
            self.callback_map.get(value, None)
            if isinstance(value, str)
            else None
        )
        if callback is None:
            return super().format_field(value, format_spec)
        else:
            return callback(value, format_spec)

    _template_cache = {}
    def load_template(self, template):
        if template not in self.__templates:
            raise ValueError("can't load templates on the fly ({} not in {})".format(
                template, self.__templates
            ))
        template = self.templates[template]
        if os.path.exists(template):
            if template not in self._template_cache:
                with open(template) as file:
                    template = file.read()
                self._template_cache[template] = template
            else:
                template = self._template_cache[template]
        return template

        # if value in s
        #     ...
        # else:
        #     super()
        # directive, args = self.parse_spec(format_spec)
        # self.directives(directive).apply(
        #     value, *args
        # )

    def vformat(self, format_string: str, args: Sequence[Any], kwargs: Mapping[str, Any]):
        try:
            self._fmt_stack.append(kwargs.copy())
            for k in self.special_callbacks:
                kwargs[k] = k
            for d in self.directives:
                if d.key+"$" not in kwargs:
                    kwargs[d.key+"$"] = d.key+"$"
            used_args = set()
            result, _ = self._vformat(format_string, args, kwargs, used_args, self.max_recusion)
            self.check_unused_args(used_args, args, kwargs)
            return result
        finally:
            self._fmt_stack.pop()

class OrderedSet(dict):
    def __init__(self, *iterable):
        if len(iterable) > 0:
            iterable = iterable[0]
        super().__init__({k:None for k in iterable})
    def union(self, other:'OrderedSet'):
        return type(self)(itertools.chain(self.keys(), other.keys()))
    def __repr__(self):
        return "{}({})".format(type(self).__name__, list(self.keys()))
    def __iter__(self):
        return iter(self.keys())
    def add(self, k):
        self[k] = None
    def update(self, ks):
        super().update({k:None for k in ks})
    def __delitem__(self, k):
        del self[k]
class Locator(typing.Protocol):
    def locate(self, identifier):
        ...
    def paths(self, **opts)->"Iterable":
        ...
class ResourcePathLocator(Locator):
    def __init__(self, path:Iterable[str]):
        if isinstance(path, str):
            path = [path]
        self.path = list(path)
    def locate(self, identifier):
        if os.path.exists(identifier):
            return os.path.abspath(identifier)
        else:
            for d in self.path:
                f = self.resource_path(d, identifier)
                if os.path.exists(f):
                    return f
    def resource_path(self, d, f):
        return os.path.join(d, f)
    def paths(self, max_depth=7, **_):
        s = OrderedSet()
        for d in self.path:
            base = self.resource_path(d, "")
            base_depth = len(pathlib.Path(base).parts)
            for root, dirs, files in os.walk(base, topdown=True):
                br = pathlib.Path(root).parts[base_depth:]
                if len(br) > max_depth:
                    break
                s.update(os.path.join(*br, f) for f in files)
        return s
    def directories(self):
        return [self.resource_path(d, "") for d in self.path]
    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            self.path
        )
class SubresourcePathLocator(ResourcePathLocator):
    def __init__(self, roots, extension):
        self.ext = extension
        super().__init__(roots)
    def resource_path(self, d, f):
        return os.path.join(d, self.ext, f)
    def __repr__(self):
        return "{}({}, {})".format(
            type(self).__name__,
            self.path,
            self.ext
        )
class ResourceLocator(Locator):
    def __init__(self,
                 locators:Iterable[Union[ResourcePathLocator,Iterable[str], Tuple[Iterable[str], Union[str, Iterable[str]]]]]
                 ):
        if isinstance(locators, str):
            locators = [locators]
        self.locators = []
        for s in locators:
            if isinstance(s, ResourcePathLocator):
                s = [s]
            elif isinstance(s[0], str):
                s = [ResourcePathLocator(s)]
            elif isinstance(s[1], str):
                s = [SubresourcePathLocator(*s)]
            else:
                s = [SubresourcePathLocator(s[0], e) for e in s[1]]
            self.locators.extend(s)
    def locate(self, identifier):
        for l in self.locators:
            res = l.locate(identifier)
            if res is not None:
                return res
    def paths(self, filter_pattern=None, **_):
        paths = {r for l in self.locators for r in l.paths()}
        if filter_pattern is not None:
            if isinstance(filter_pattern, str):
                try:
                    filter_pattern = re.compile(filter_pattern)
                except (re.error, ValueError):
                    filter_pattern = re.compile(fnmatch.translate(filter_pattern))
            paths = OrderedSet(p for p in paths if filter_pattern.match(p))
        return paths
    def directories(self):
        return OrderedSet(r for l in self.locators for r in l.directories())
    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            self.directories()
        )

class TemplateEngine:
    """
    Provides an engine for generating content using a
    `TemplateFormatter` and `ResourceLocator`
    """

    formatter_class = TemplateFormatter
    def __init__(self,
                 locator:Locator,
                 template_pattern="*.*",
                 ignore_missing=False,
                 formatter_class=None,
                 ignore_paths=()
                 ):
        self.locator = locator
        self.templates = {
            k:self.locator.locate(k)
            for k in self.locator.paths(filter_pattern=template_pattern)
        }
        if formatter_class is None:
            formatter_class = self.formatter_class
        self.formatter = formatter_class(self.templates)
        self.ignore_missing = ignore_missing
        self.ignore_paths = ignore_paths
    def __repr__(self):
        cls = type(self)
        return dev.str_elide(f"{cls.__name__}({self.locator})", width=50)


    def format_map(self, template, parameters):
        if template in self.templates:
            template = self.formatter.load_template(template)
        return self.formatter.vformat(
            template,
            (),
            parameters if not self.ignore_missing else collections.defaultdict(parameters)
        )
    def format(self, template, **parameters):
        return self.format_map(template, parameters)

    class outStream:
        def __init__(self, file, mode='w+', **kw):
            self.file = file
            self.file_handle = None
            self.mode = mode
            self.kw = kw

        def __enter__(self):
            if self.file_handle is None:
                if isinstance(self.file, str):
                    try:
                        os.makedirs(os.path.dirname(self.file))
                    except OSError:
                        pass
                    self.file_handle = open(self.file, self.mode, **self.kw)
                else:
                    self.file_handle = self.file
            return self.file_handle

        def __exit__(self, exc_type, exc_val, exc_tb):
            if isinstance(self.file, str):
                self.file_handle.close()
            self.file_handle = None

        def write(self, s):
            with self as out:
                out.write(s)
            return self.file
    def write_string(self, target, txt):
        return self.outStream(target).write(txt)
    def apply(self, template, target, **template_params):
        try:
            if target is None:
                return self.format_map(template, template_params)
            elif target not in self.ignore_paths:
                return self.write_string(target, self.format_map(template, template_params))
        except:
            raise ValueError("{}: error in filling template {}".format(
                type(self).__name__,
                template
            ))

class TemplateHandler(ObjectHandler):
    template = None
    extension = ".md"
    squash_repeat_packages=True
    def __init__(self,
                 obj,
                 *,
                 out=None,
                 engine:TemplateEngine=None,
                 root=None,
                 squash_repeat_packages=True,
                 is_package_root=False,
                 **extra_fields
                 ):
        super().__init__(obj, **extra_fields)
        self.is_package_root = is_package_root
        self.squash_repeat_packages = squash_repeat_packages
        self.target = self.get_output_file(out)
        # if root is None:
        #     root = os.path.dirname(self.target)
        if root is None and out is not None and os.path.isdir(out):
            root = out
        self.root = root
        if engine is None:
            raise ValueError("{}:`engine` can't be None".format(type(self).__name__))
        self.engine = engine

        test_path = os.path.join(*self.identifier.split(".")) + ".md"
        if self.engine.locator.locate(test_path) is not None:
            self.template = test_path
        if self.template is None:
            raise ValueError("{}: template can't be None".format(type(self).__name__))

    def template_params(self, **kwargs):
        base_parms = self.extra_fields.copy()
        base_parms.update(self.get_template_params(**kwargs))
        return base_parms
    @abc.abstractmethod
    def get_template_params(self, **kwargs):
        """
        Returns the parameters that should be inserted into the template

        :return:
        :rtype: dict
        """
        raise NotImplementedError("abstract base class")

    @property
    def package_path(self):
        return self.get_package_and_url()
    def get_package_and_url(self, include_url_base=True):
        """
        Returns package name and corresponding URL for the object
        being documented
        :return:
        :rtype:
        """
        ident = self.identifier
        if self.squash_repeat_packages:
            ident = self.squash_reps(ident)
        pkg_split = ident.split(".", 1)

        if len(pkg_split) == 1:
            pkg = pkg_split[0]
            rest = ""
        elif len(pkg_split) == 0:
            pkg = ""
            rest = "Not.A.Real.Package"
        else:
            pkg, rest = pkg_split

        # if self.squash_repeat_packages:
        #     base_id = rest.split(".")
        #     new_id = [base_id[0]]
        #     for i,k in enumerate(base_id[1:]):
        #         if new_id[-1] != k:
        #             new_id.extend(base_id[1+i:])
        #             break
        #     rest = ".".join(new_id)

        # if len(rest) == 0:
        #     file_url = "__init__.py"
        # else:
        file_url = ident.replace(".", "/")
        if self.is_package_root:
            file_url += "/__init__.py"
        else:
            file_url += '.py'

        if include_url_base and 'url_base' in self.extra_fields:
            file_url = self.extra_fields['url_base'] + "/" + file_url
        return pkg, file_url

    @property
    def target_identifier(self):
        return ".".join(self.get_target_extension())
    @classmethod
    def squash_reps(cls, ident):
        ident = ident.split('.')
        up = ident[0]
        i = -1
        for i,r in enumerate(ident[1:]):
            if r != up:
                i = i - 1
                break
        return '.'.join(ident[i+1:])
    def get_target_extension(self, identifier=None):
        if identifier is None:
            identifier = self.identifier
        elif not isinstance(identifier, str):
            identifier = self.get_identifier(identifier)
        base_id = identifier.split(".")
        if self.squash_repeat_packages:
            new_id = [base_id[0]]
            for i, k in enumerate(base_id[1:]):
                if new_id[-1] != k:
                    new_id.extend(base_id[1 + i:])
                    break
            base_id = new_id
        return base_id
    def get_output_file(self, out):
        """
        Returns package name and corresponding URL for the object
        being documented
        :return:
        :rtype:
        """
        if out is None:
            out = sys.stdout
        elif isinstance(out, str):# and os.path.isdir(out):
            base_id = self.identifier.split(".")
            if self.squash_repeat_packages:
                new_id = [base_id[0]]
                for i, k in enumerate(base_id[1:]):
                    if new_id[-1] != k:
                        new_id.extend(base_id[1 + i:])
                        break
                base_id = new_id
            out = os.path.join(out, *base_id) + self.extension
        return out
    def handle(self, template=None, target=None, write=True):
        """
        Formats the documentation Markdown from the supplied template

        :param template:
        :type template:
        :return:
        :rtype:
        """
        if self.check_should_write():
            if template is None:
                template = self.template
            params = self.template_params()
            if target is None:
                out_file = self.target
            else:
                out_file = target
            if isinstance(out_file, str):
                pkg, file_url = self.package_path
                params['package_name'] = pkg
                params['file_url'] = file_url
                params['package_url'] = os.path.dirname(file_url)

                if self.root is not None:
                    root = self.root
                    root_split = pathlib.Path(root).parts
                    out_split = pathlib.Path(out_file).parts
                    root_depth = len(root_split)
                    out_url = "/".join(out_split[root_depth:])
                else:
                    file_split = pathlib.Path(file_url).parts
                    out_split = pathlib.Path(out_file).parts
                    out_url = "/".join(out_split[-len(file_split):])
                params['file'] = out_file
                params['url'] = out_url
            try:
                out = self.engine.apply(template, out_file if write else None, _self=self, **params)
            except KeyError as e:
                raise ValueError("{} ({}): template needs key {}".format(
                    type(self).__name__,
                    self.obj,
                    e.args[0]
                ))
            except IndexError as e:
                raise ValueError("{} ({}): template index {} out of range...".format(
                    type(self).__name__,
                    self.obj,
                    e.args[0]
                ))
            except Exception as e:
                raise ValueError("{} ({}): {}".format(
                    type(self).__name__,
                    self.obj,
                    self.template,
                    e
                ))
            return out
        else:
            if not write:
                return ""

    blacklist_packages = {'numpy', 'scipy', 'matplotlib'} #TODO: more sophisticated blacklisting
    def check_should_write(self):
        """
        Determines whether the object really actually should be
        documented (quite permissive)
        :return:
        :rtype:
        """
        base = self.identifier.split(".", 1)[0]
        stdlib = base == 'builtins'
        if not stdlib:
            try:
                loc = sys.modules[base].__file__
            except KeyError:
                stdlib = False
            else:
                stdlib = loc.startswith(sys.prefix) and 'site-packages' not in loc
        return not stdlib and base not in self.blacklist_packages

class TemplateResourceExtractor(ResourceLocator):
    extension = '.md'
    def path_extension(self, handler:TemplateHandler):
        """
        Provides the default examples path for the object
        :return:
        :rtype:
        """
        return os.path.join(*handler.get_target_extension()) + self.extension
    resource_keys = []
    resource_attrs = []
    def get_resource(self, handler:TemplateHandler, keys=None, attrs=None):
        if keys is None:
            keys = self.resource_keys
        if attrs is None:
            attrs = self.resource_attrs
        for k in keys:
            res = handler[k]
            if res is not None:
                res_file = self.locate(res)
                if res_file is not None:
                    res = res
                break
        else:
            for a in attrs:
                res = getattr(handler.obj, a, None)
                if res is not None:
                    res_file = self.locate(res)
                    if res_file is not None:
                        res = res
                    break
            else:
                ext = self.path_extension(handler)
                if isinstance(ext, str):
                    res = self.locate(ext)
                else:
                    for e in ext:
                        res = self.locate(e)
                        if res is not None:
                            break
                    else:
                        res = None
        return res
    def load(self, handler:TemplateHandler):
        """
        Loads examples for the stored object if provided
        :return:
        :rtype:
        """

        resource = self.get_resource(handler)
        if resource is not None and os.path.isfile(resource):
            with open(resource) as f:
                resource = f.read()
        return resource

class ModuleTemplateHandler(TemplateHandler):
    template = 'module.md'
class ClassTemplateHandler(TemplateHandler):
    template = 'class.md'
class FunctionTemplateHandler(TemplateHandler):
    template = 'function.md'
class MethodTemplateHandler(TemplateHandler):
    template = 'method.md'
class ObjectTemplateHandler(TemplateHandler):
    template = 'object.md'
class IndexTemplateHandler(TemplateHandler):
    template = 'index.md'

class TemplateWalker(ObjectWalker):
    module_handler = ModuleTemplateHandler
    class_handler = ClassTemplateHandler
    function_handler = FunctionTemplateHandler
    method_handler = MethodTemplateHandler
    object_handler = ObjectTemplateHandler
    index_handler = IndexTemplateHandler
    def __init__(self, engine:TemplateEngine, out=None, description=None, **extra_fields):
        self.engine = engine
        self.out_dir = out
        self.description = description
        super().__init__(**extra_fields)

    @property
    def default_handlers(self):
        return collections.OrderedDict((
            ((str, types.ModuleType), self.module_handler),
            ((type,), self.class_handler),
            ((types.FunctionType,), self.function_handler),
            (None, self.object_handler)
        ))

    def get_handler(self, obj, *, out=None, engine=None, tree=None, **kwargs):
        return super().get_handler(
            obj,
            out=self.out_dir,
            engine=self.engine,
            tree=tree,
            **kwargs
        )

    def visit_root(self, o, **kwargs): # here for overloading
        return self.visit(o, **kwargs)

    def write(self, objects, max_depth=-1, index='index.md'):
        """
        Walks through the objects supplied and applies the appropriate templates
        :return: index of written files
        :rtype: str
        """

        if self.out_dir is not None and index is not None:
            try:
                os.makedirs(self.out_dir)
            except OSError:
                pass
            out_file = os.path.join(self.out_dir, index)
        else:
            out_file = None

        files = [
            self.visit_root(o, max_depth=max_depth)
            for o in objects
        ]
        files = [f for f in files if f is not None]
        w = self.get_handler(files,
                             cls=self.index_handler,
                             out=out_file,
                             engine=self.engine,
                             root=self.out_dir,
                             extra_fields=self.extra_fields,
                             description=self.description
                             )
        return w.handle()

class TemplateResourceList(Locator):
    """
    Implements the `ResourceLocator` interface, but is backed by a `dict` of
    explicit resources rather than a set of paths.
    """
    def __init__(self, resource_dict:Mapping[str, Any]):
        self.dict = resource_dict
    def paths(self, **_):
        return self.dict.keys()
    def locate(self, identifier):
        return self.dict.get(identifier, None)
class TemplateInterfaceList(TemplateResourceList):
    """
    A set of functions to be used to construct interfaces
    """
    def __init__(self, resource_dict:Mapping[str, Callable]):
        super().__init__(resource_dict)
class TemplateInterfaceFormatter:
    """
    Provides an interface that mimics a `TemplateFormatter`
    but does nothing more than route to a set of template functions
    """
    def __init__(self, templates):
        self.__templates = templates
        self._fmt_stack = []
    @property
    def format_parameters(self):
        return self._fmt_stack[-1] if len(self._fmt_stack) > 0 else None
    @property
    def templates(self):
        return self.__templates
    @property
    def special_callbacks(self):
        return {
            # "%":self.apply_eval_tree,
            # "$":self.apply_directive_tree,
            # "#":self.apply_comment,
            # 'raw$':self.apply_raw,
            # 'assign%':self.apply_assignment
        }
    # @property
    # def callback_map(self):
    #     return dict(
    #         self.special_callbacks,
    #         **{d.key+"$":self.apply_directive for d in self.directives}
    #     )
    #
    # def apply_eval_tree(self, _, spec) -> str:
    #     tree = ast.parse(inspect.cleandoc(spec))
    #     ev = TemplateASTEvaluator(self, self.directives, self.format_parameters).evaluate_node(tree)
    #     if ev is None:
    #         ev = ""
    #     return ev
    # def apply_directive_tree(self, _, spec) -> str:
    #     return self.apply_eval_tree(_, "("+spec+")")
    # def apply_assignment(self, key, spec) -> str:
    #     key, val = spec.split("=", 1)
    #     self.format_parameters[key] = val
    #     return ""
    # def apply_raw(self, key, spec) -> str:
    #     return spec
    # def apply_comment(self, key, spec) -> str:
    #     return ""
    # def apply_directive(self, key, spec) -> str:
    #     return self.apply_directive_tree(
    #         key,
    #         "{}({})".format(key.strip("$"), spec)
    #     )
    # def format_field(self, value: Any, format_spec: str) -> str:
    #     if self.format_parameters is None:
    #         raise NotImplementedError("{}.{} called outside of `vformat`".format(
    #             type(self).__name__,
    #             'format_field'
    #         ))
    #     callback = (
    #         self.callback_map.get(value, None)
    #         if isinstance(value, str)
    #         else None
    #     )
    #     if callback is None:
    #         return super().format_field(value, format_spec)
    #     else:
    #         return callback(value, format_spec)

    # _template_cache = {}
    def load_template(self, template):
        return self.templates[template]

        # if value in s
        #     ...
        # else:
        #     super()
        # directive, args = self.parse_spec(format_spec)
        # self.directives(directive).apply(
        #     value, *args
        # )

    def vformat(self, template: Callable, args: Sequence[Any], kwargs: Mapping[str, Any]):
        try:
            self._fmt_stack.append(kwargs.copy())
            for k in self.special_callbacks:
                kwargs[k] = k
            return template(*args, **kwargs)
            # for d in self.directives:
            #     if d.key+"$" not in kwargs:
            #         kwargs[d.key+"$"] = d.key+"$"
            # used_args = set()
            # result, _ = self._vformat(format_string, args, kwargs, used_args, self.max_recusion)
            # self.check_unused_args(used_args, args, kwargs)
            # return result
        finally:
            self._fmt_stack.pop()

class TemplateInterfaceEngine(TemplateEngine):
    """
    A variant on a template engine designed for more interactive use.
    In many ways, _not_ a template engine, but too useful to ignore while I
    find a more uniform abstraction.
    Generates _interfaces_ from a set of interface template functions
    rather than strings from template files.
    """

    formatter_class = TemplateInterfaceFormatter
    def __init__(self,
                 templates: 'TemplateInterfaceList|dict',
                 ignore_missing=False,
                 formatter_class=None,
                 ignore_paths=()
                 ):
        if isinstance(templates, dict):
            templates = TemplateInterfaceList(templates)
        super().__init__(
            templates,
            template_pattern=None,
            ignore_missing=ignore_missing,
            formatter_class=formatter_class,
            ignore_paths=ignore_paths
        )

    def format_map(self, template, parameters):
        if template in self.templates:
            template = self.formatter.load_template(template)
        return self.formatter.vformat(
            template,
            (),
            parameters if not self.ignore_missing else collections.defaultdict(parameters)
        )

    def apply(self, template, target, **template_params):
        try:
            if target is None or target is sys.stdout:
                return self.format_map(template, template_params)
            elif target not in self.ignore_paths:
                return {target:self.format_map(template, template_params)}
        except:
            raise ValueError("{}: error in filling template {}".format(
                type(self).__name__,
                template
            ))


import numpy as np, re
from collections import OrderedDict
from .RegexPatterns import *
from .StructuredType import *

__all__ = [
    "StringParser",
    "StringParserException"
]

__reload_hook__ = ['.RegexPatterns']

########################################################################################################################
#
#                                           StringParser
#
class StringParserException(Exception):
    ...

# """
# One thing that's really tough with this stuff is figuring out how to manage shapes as we move across the different parsers and recursion level
#
# We can use the RegexPattern to manage the original StructuredType -> this seems to be pretty reliable
# We can use that to create an initial StructuredTypeArray -> this seems to be pretty reliable
# Then if we call parse_all, we need to reshape that StructuredTypeArray so that it can take a vector of return values
#     -> I _think_ this works as it should
# Next, no matter which parser we use, we need to get out our match or matches -> this is, again, quite reliable
# At that point we need to figure out how many groups each subarray should have -> this seems to be broken
#
#
# """
#
# """
# StringParser Design:
#
# input: RegexPattern
# output: StructuredTypeArray or list or OrderedDict thereof
#
# Basic Idea:
#
# RegexPattern can track which of its children are tagged as Capturing and which are tagged as Named. These children in turn
# can keep track of their data type (dtype) and also the shape of the data that should come out of their matched groups.
#
# So once we've compiled down RegexPattern--making sure that only groups at the outer-most level capture when compiling--
# what we can do is use the python re module to find the various matched components.
#
# At this point, we need to convert the matched strings into the appropriately formatted NumPy arrays. To handle this in
# recursive fashion I created a StructuredTypeArray class that has two flavors, simple and compound, which respectively
# handle the direct communication with NumPy and the delegation and distribution of higher-order data to simpler arrays.
#
# So in our parsing process we need to create a StructuredTypeArray for the found dtype from the RegexPattern. This is
# relatively straightforward. Next, we might be using parse_all which asks for a vector of results back. If this is the
# case we need to take our initial StructuredTypeArray and turn it into, in effect, a StructuredTypeArray vector. Rather
# than use an inefficient python list format, though, we call StructuredTypeArray.add_axis which broadcasts to a
# higher-dimensional data format. It also changes up the stored StructuredType such that a new axis has been prepended to
# the shapes of each.
#
# Next we have to work with our matched data somehow. There are a few issues here, but probably the biggest one is that the
# number of matched groups doesn't need to line up with the number of blocks we want returned. This means that we might need to take our
# matched groups and use some 'block_size' that each StructuredTypeArray tracks to figure out how to split these groups up.
#
# Assuming that works well, we now apply the block handler for each declared Capturing block to the matched data string.
# For very simple data, this will be effectively doing nothing, so that the StructuredTypeArray can handle the cast to an
# np.ndarray. For compound types or more complex data we need to basically go match-by-match and apply the block parser.
#
# This is where the recursion occurs and where things can get messy. The default block handler for an np.ndarray calls
# parse on a string. This means that parse will be called until the data can finally be put into a StructuredTypeArray
# format. On the other hand, the returned data needs to be consistent with the expected return type of the parent call.
# For that reason, the block handler passes in the return array, and the return array uses its stype when calling the
# subparser. This means, though, that there could be an apparent shape mismatch if the data itself is ragged. To resolve
# that, StructuredTypeArray will pad lower-dimensional data into a higher-dimensional format when possible.
#
# A major complicating factor in all of this is how one manages Repeating patterns. Every Repeating pattern actually needs
# to be managed using parse_all and its own Capturing groups, after first finding the matches for the repeats. Since the
# passed dtype for such a call is also a higher-dimensional object, the outer-most axis of the dtype needs to be dropped,
# then parse_all needs to be called on the child form. Managing this correctly inside a recursive process can be challenging.
#
# """
class StringParser:
    """
    A convenience class that makes it easy to pull blocks out of strings and whatnot
    """

    def __init__(self, regex:RegexPattern):
        self.regex = regex

    def parse(self,
              txt,
              regex=None,
              block_handlers=None,
              dtypes=None,
              out=None
              ):
        """Finds a single match for the and applies parsers for the specified regex in txt

        :param txt: a chunk of text to be matched
        :type txt: str
        :param regex: the regex to match in _txt_
        :type regex: RegexPattern
        :param block_handlers: handlers for the matched blocks in _regex_ -- usually comes from _regex_
        :type block_handlers: iterable[callable] | OrderedDict[str: callable]
        :param dtypes: the types of the data that we expect to match -- usually comes from _regex_
        :type dtypes: iterable[type | StructuredType] | OrderedDict[str: type | StructuredType]
        :param out: where to place the parsed out data -- usually comes from _regex_
        :type out: None | StructuredTypeArray | iterable[StructuredTypeArray] | OrderedDict[str: StructuredTypeArray]
        :return:
        :rtype:
        """

        # we'll enforce that the return type be a StructuredTypeArray
        # the enforcement mechanism here will be basically to take the Regex dtype if provided and otherwise
        # to coerce to a StructuredType

        if regex is None:
            regex = self.regex

        if isinstance(regex, RegexPattern) and regex.parser is not None:
            return regex.parser(txt)

        # one lingering question is how to handle `Repeating` patterns.
        # Probably want to extract the child block and feed that into parse_all...
        # might need to figure out some alternate manner for handling dtypes and block_handlers
        # in this case since I think the delegation to the parse_all might fuck stuff up...
        if isinstance(regex, RegexPattern) and regex.is_repeating:
            import copy

            iterable_regex = copy.copy(regex)
            iterable_regex.pat = lambda p, *a, **kw: p # to unbind the Repeating capture screwups
            parser = type(self)(iterable_regex)

            # ah fack how do we handle the dtypes passing?
            # if dtypes was passed...I guess we gotta feed it through, right?
            # Or well maybe not...? If it was passed it should be a _vector_ stype, right?
            # So I guess we gotta strip the outer level of repeats...?
            if out is None:
                if isinstance(dtypes, StructuredType):
                    dtypes = dtypes.drop_axis()
                elif dtypes is not None:
                    dtypes = [d.drop_axis() for d in dtypes]
            else:
                dtypes = None
            if isinstance(out, (dict, OrderedDict)):
                out = out.copy()
                out['single'] = True
            else:
                out = {'array':out, 'single':True}

            res = parser.parse_all(txt, dtypes=dtypes, out=out)

        else:
            if block_handlers is None and isinstance(regex, RegexPattern):
                block_handlers = self.get_regex_block_handlers(regex)
            if dtypes is None and isinstance(regex, RegexPattern):
                dtypes = self.get_regex_dtypes(regex)

            if out is None:
                res = self._set_up_result_arrays(dtypes)
                append = False
                single = None
            elif isinstance(out, (dict, OrderedDict)):
                res = out['array']
                try:
                    append = out['append']
                except KeyError:
                    append = False
                try:
                    single = out['single']
                except KeyError:
                    single = None
            else:
                res = out
                append = False
                single = None

            # if we're given a non-string regex we compile it
            pattern = None
            if isinstance(regex, RegexPattern):
                pattern = regex
                regex = regex.compiled

            match = re.search(regex, txt)
            try:
                self._handle_parse_match(match, res, block_handlers, append=append, single=single)
            except Exception as e:
                if append or (isinstance(append, int) and append == 0):
                    raise StringParserException('failed to append to results array {} for match {}'.format(
                        res,
                        match
                    )) from e
                else:
                    raise StringParserException('failed to insert into results array {} for match {}'.format(
                        res,
                        match
                    )) from e

        # one issue here is that by default this would return a list or some structured object like that
        # sometimes we don't _want_ that list
        # if we have something like Repeating(Capturing(Number), ...)
        # we don't really want [ StructuredTypeArray(...) ] returned
        # what we'd really like to get back is just the StructuredTypeArray
        # now here's the question, is there a case where if we have [ StructuredTypeArray ]
        # that we _wouldn't_ just want the array? My sense is no since StructuredType should simplify the
        # singular type by default...?
        # Or will this be unique to Repeating patterns?

        # Eh, we'll assume this is mis-structured since we've taken a basic RegexPattern and
        # then captured a repeating portion of it or some set of repeating parts of it but
        # didn't bother to simplify to just that Capturing part of the RegexPattern...?
        if not isinstance(res, (OrderedDict, StructuredTypeArray)) and len(res) == 1:
            res = res[0]

        return res

    # we need to add some stuff to generalize out the parser, add some useful general methods
    # like the handling of blocks of unknown size / unknown numbers of atoms / efficient casting from string
    # to NumPy types
    def parse_all(self,
                  txt,
                  regex = None,
                  num_results = None,
                  block_handlers = None,
                  dtypes = None,
                  out = None
                  ):

        if regex is None:
            regex = self.regex

        if block_handlers is None and isinstance(regex, RegexPattern):
            block_handlers = self.get_regex_block_handlers(regex)
        if dtypes is None and isinstance(regex, RegexPattern):
            dtypes = self.get_regex_dtypes(regex)

        if out is None:
            res = self._set_up_result_arrays(dtypes)
            if isinstance(res, StructuredTypeArray):
                res.add_axis()
            elif isinstance(res, (dict, OrderedDict)):
                for r in res.values():
                    r.add_axis()
            else:
                for r in res:
                    r.add_axis()
            append = False
            single = None
        elif isinstance(out, (dict, OrderedDict)):
            res = out['array']
            try:
                append = out['append']
            except KeyError:
                append = False
            try:
                single = out['single']
            except KeyError:
                single = None
        else:
            res = out
            append = False
            single = None

        # if we're given a non-string regex we compile it
        base_regex = regex
        # print(">>>>>>>>>", type(regex.compiled))
        if isinstance(regex, RegexPattern):
            # print("...wat?", type(regex.compiled))
            regex = regex.compiled

        try:
            match_iter = re.finditer(regex, txt)
        except:

            raise Exception(type(regex), type(regex.compiled))
        if num_results is not None:
            import itertools as it
            matches = list(it.islice(match_iter, None, num_results))
        else:
            matches = list(match_iter)

        if matches is None or len(matches)==0:
            raise StringParserException("Pattern {} not found in {}".format(regex, txt))

        self._handle_parse_match(matches, res, block_handlers, append = append, single = single)
        #  we don't need to append since we've pushed the filling iteration into the StructuredArray

        return res

    class MatchIterator:
        class Match:
            def __init__(self, parent, block):
                self.parent = parent
                self.block = block
            @property
            def value(self):
                res = self.parent.parser._set_up_result_arrays(self.parent.dtypes)
                self.parent.parser._handle_parse_match(self.block, res, block_handlers=self.parent.block_handlers)
                return res

        def __init__(self, parser, match_iter, num_results, dtypes, block_handlers):
            if num_results is not None:
                import itertools as it
                match_iter = it.islice(match_iter, None, num_results)

            self.parser = parser
            self.dtypes = dtypes
            self.block_handlers = block_handlers
            self.match_iter = match_iter

        def __iter__(self):
            for match in self.match_iter:
                yield self.Match(self, match)
        def __next__(self):
            match = next(self.match_iter)
            return self.Match(self, match)

    def parse_iter(self, txt,
                  regex = None,
                  num_results = None,
                  block_handlers = None,
                  dtypes = None
                  ):

        if regex is None:
            regex = self.regex

        if block_handlers is None and isinstance(regex, RegexPattern):
            block_handlers = self.get_regex_block_handlers(regex)
        if dtypes is None and isinstance(regex, RegexPattern):
            dtypes = self.get_regex_dtypes(regex)

        # if we're given a non-string regex we compile it
        if isinstance(regex, RegexPattern):
            regex = regex.compiled

        match_iter = re.finditer(regex, txt)
        return self.MatchIterator(self, match_iter, num_results, dtypes, block_handlers)


    #region RegexPattern data extractors
    @classmethod
    def get_regex_block_handlers(cls, regex):
        """Uses the uncompiled RegexPattern to determine what blocks exist and what handlers they should use

        :param regex:
        :type regex: RegexPattern
        :return:
        :rtype:
        """
        # I think it might make sense to allow a given regex to declare its _own_ parser that specifies how it
        # should be used in parse and parse_all. This can remain the default implementation, but that would allow
        # one to push the complexity of figuring out how to handle an object onto the object itself...

        # since we handle repeating patterns by delegating to parse_all by default, these handlers will need to be the
        # handlers for the child parse_all call instead of the basic parser from the first matched child, I think...
        # ah actually more generally we need to make sure that for each captured block (i.e. what powers parser)
        # we have a handler
        # the basic assumption has generally been that all children will be relevant, but as things
        # increase in complexity this assumption breaks down
        # therefore the regex needs to remember itself which of its children are capturing and which aren't

        handlers = regex.named_groups
        if handlers is not None:
            for k,r in tuple(handlers.items()): # we extract the full list first so we can go back and edit the handlers map
                handlers[k] = cls._get_regex_handler(r)
        else:
            # handlers = None
            # if we have no named groups to work with we'll create a block handler for the capturing groups?

            handlers = regex.capturing_groups
            if handlers is not None:
                if isinstance(handlers, RegexPattern):
                    handlers = [ handlers ]

                dtype = regex.dtype
                if dtype.shape is None:
                    handlers = None
                elif dtype.is_simple and len(handlers) == dtype.shape[-1]:
                    handlers = [ None ] # means we _wanted_ to have one block, but we had to declare it over multiple CapturingGroups
                else:
                    handlers = [ cls._get_regex_handler(r) for r in handlers if r.dtype is not DisappearingType ]


        return handlers
    @classmethod
    def get_regex_dtypes(cls, regex):
        """Uses the uncompiled RegexPattern to determine which StructuredTypes to return

        :param regex:
        :type regex: RegexPattern
        :return:
        :rtype:
        """

        dtypes = regex.dtype
        # print(">"*50, dtypes)
        # if dtypes is not None:
        #     for k, r in tuple(dtypes.items()):
        #         dtypes[k] = r.dtype
        # else:## there is some question as to whether I might actually want to make it so the _children_ get parsed out...
        #     dtypes = regex.capturing_groups
        #     if dtypes is not None:
        #         if isinstance(dtypes, RegexPattern):
        #             dtypes = dtypes.dtype
        #         else:
        #             dtypes = tuple(d.dtype for d in dtypes)
        #
        #     else:
        #         dtypes = regex.dtype
        # print(dtypes, "<"*50, sep="\n")
        return dtypes
    #endregion

    #region StructuredTypeArray interface

    def _set_up_result_arrays(self, dtypes):
        # we'll always force our dtypes and results to be a StructuredType and StructuredTypeArray
        # as these will force the results to be conformaing
        res = StructuredTypeArray(dtypes)
        return res

    #endregion

    #region Match Handler
    def _split_match_groups(self, groups, handlers, res):
        """This is _supposed_ to be splitting my groups so that the appropriate handler handles them, but it's missing a corner case
        Sometimes we have something like:
            RegexPattern((Capturing(Number), ..., Capturing(Number), ..., Capturing(Number))

        This happens to make 3 handlers even though it's really a single array of data...
        Then the grouping sees this and says "Aha! Must transpose"
        In reality it shouldn't even be bothering to touch it though...

        :param groups:
        :type groups:
        :param handlers:
        :type handlers:
        :param res:
        :type res:
        :return:
        :rtype:
        """

        if handlers is None:
            gg = groups
        elif len(groups[0]) == len(handlers): # the is None clause might be unnecessary...
            # just need to transpose the groups, basically,
            gg = groups.T
        elif len(groups[0]) < len(handlers):
            raise StringParserException(
                "{}.{}: got {} block handlers but only {} matched elements".format(
                    type(self).__name__,
                    '_split_match_groups',
                    len(handlers),
                    len(groups[0])
                )
            )
        elif all(isinstance(r, StructuredTypeArray) for r in res):
            # we have _more_ groups than handlers so we need to chunk them up
            blocks = [ r.block_size for r in res ]
            gg = [ None ] * len(blocks)
            sliced = 0
            for i, b in enumerate(blocks):
                if b == 1:
                    gg[i] = groups[:, sliced]
                else:
                    gg[i] = groups[:, sliced:sliced+b]
                sliced += b
        elif len(handlers) == 1:
            gg = [groups]
        else:
            gg = groups
        # print(">"*50)
        # print("Got in groups:", groups)
        # print("Got out:", gg)
        # print("<"*50)
        return gg # groups should now be chunked into groups if it wasn't before

    def _handle_insert_result(self, array, handler, data, single=True, append=False):
        """Handles the process of actually inserting the matches data/groups into array, applying any necessary
        handler in the process

        :param array: the array the data should be inserted into
        :type array: StructuredTypeArray
        :param handler: handler function for further processing of the matched data
        :type handler: None | callable
        :param data: matched data
        :type data: iterable[str]
        :param single: whether data should be treated as a single object (e.g. a subarray) or as multiple data objects
        :type single: bool
        :param append: whether the data should be appended or whether the array should entirely re-assign itself / which axes to assign to
        :type append: bool | iterable[int]
        """
        if handler is not None:
            data = handler(data, array=array, append=append)

        if data is not None:
            if array.has_indeterminate_shape:
                array.fill(data)
            elif single and (append or (append is not False and isinstance(append, int) and append == 0)):
                # print("pre-append:>", array)
                # print("axis:", append)
                # print("total axis:", append + array.append_depth)
                # print(data)
                if append is True or not isinstance(append, int):
                    append = 0
                array.append(data, axis=append)
                # print("<:post-append", array)
            elif (append or (append is not False and  isinstance(append, int) and append == 0)):
                # print("pre-extend:>", array)
                if append is True or not isinstance(append, int):
                    append = 0
                array.extend(data, single=single, axis=append)
                # print("<:post-extend", array)
            else:
                # print("pre-fill:>", array)
                array.fill(data)
                # print("<:post-fill", array)
    def _handle_parse_match(self, match, res,
                            block_handlers,
                            append=False,
                            single=None,
                            default_value = ""
                            ):
        """Figures out how to handle the matched data from the parser
        In general the way we _want_ to do this is to basically do the most minimal amount of processing we can
        and then use the efficient methods inside StructuredTypeArray to cast this to properly typed arrays

        One issue is that we can end up with stuff that looks like:
            [
                [ group_a match_1, group_a match_2, group_b match_1, ... ]
                ...
            ]
        where we need to first chop the data into group_a, group_b, ...

        The type-calculus in RegexPattern and StructuredType can give us a "block size" for each subarray inside the
        StructuredTypeArray which indicates how many elements it expects to see from the group

        We use that to chop up the array as needed in these multi-match cases before trying to feed it into the
        StructuredTypeArray

        Complicating that, though, is that we sometimes need to do an extra layer of parsing on our matches

        Imagine that we found every instance of Repeating(Capturing(Number), suffix=Optional(Whitespace))

        This could give us something like
            [
                '-1232.123-234.345 3434.3434 10000000000',
                '5',
                '6'
            ]

        It would be hard to say what we should do with this, so we'll assume (possibly incorrectly) that we won't get a
        ragged array. If we need to introduce raggedness we'll do it later

        Even then, we could have
            [
                '-1232.123-234.345 3434.3434 10000000000',
                '5         5       5         5',
                '6         6       6         6'
            ]

        Here we'd need to go through line-by-line and apply the subparser of Number to each line to split it up

        On the other hand, what if we matched Duplicated(Capturing(Number), 4, riffle=Optional(Whitespace))?

        Then we'd have
            [
                ['-1232.123', '-234.345', '3434.3434', '10000000000'],
                ['5',         '5',        '5',         '5'],
                ['6',         '6',        '6',         '6']
            ]

        And here we wouldn't want to apply the subparser.

        The only way we could determine the difference is by comparing the shape of the StructuredTypeArray to the
        matched data. This means our block_handlers need to take the passed res array to figure their shit out too.


        :param append: whether to append the data or not
        :type append:
        :param single: whether the matched data should be treated like a singular object
        :type single:
        :param default_value:
        :type default_value:
        :param match:
        :type match: re.Match | iterable[re.Match]
        :param res:
        :type res:
        :param block_handlers:
        :type block_handlers:
        :return:
        :rtype:
        """

        if hasattr(match, 'groupdict'):
            single_match = True
            if single is None:
                single = single_match
            first_match = match # indicates we were fed a single match
        else:
            single_match = False
            if single is None:
                single = single_match
            if match is None or len(match) == 0:
                raise StringParserException("Can't handle series of matches if there are no matches")
            first_match = match[0] # indicates we were fed a series of matches
            # indicates we were fed a single group

        gd = first_match.groupdict(default=default_value) if res.dict_like else []
        # print("WTF", first_match, first_match.groups())
        if len(gd) == 0:
            # we have no _named_ groups so we just pull everything and run with it

            # one difficulty about this setup is that it makes vector data into a 2D array
            # on the other hand I don't think there's anything else to be done here?
            groups = match.groups(default=default_value) if (single_match) else np.array([ m.groups(default=default_value) for m in match ], dtype=str)

            # from here we might need to be a bit clever, actually...
            # the number of groups should align with the number of block_handlers, of course, which _should_ be connected
            # to what res looks like

            if not single_match:
                groups = self._split_match_groups(groups, block_handlers, res._array)

            if block_handlers is None:
                # print("-"*10 + "No Handler" + "-"*10)
                self._handle_insert_result(res, None, groups, single = single, append = append)
            elif len(block_handlers) == 1:
                # print("-"*10 + "Single Handler" + "-"*10)
                handler = block_handlers[0]
                self._handle_insert_result(res, handler, groups[0], single = single, append = append)
            # elif len(block_handlers) > :
            #     raise StringParserException("{}.{}: got multiple block_handlers {} but expected a singular dtype".format(
            #         type(self).__name__,
            #         '_handle_parse_match',
            #         block_handlers
            #     ))
            elif len(block_handlers) == 0:
                raise StringParserException("No block handlers available to parse data")
            else:
                # print("-"*10 + "Multiple Handlers" + "-"*10)
                # print(block_handlers, res, groups)
                for b,a,g in zip(block_handlers, res._array, groups):
                    self._handle_insert_result(a, b, g, single=single, append=append)

        else:
            # we have named groups so we only use the data associated with them
            if block_handlers is None:
                block_handlers = {}
            groups = gd if single_match else [ m.groupdict(default=default_value) for m in match ]
            # now we need to remerge the groupdicts by name if not single
            if not single_match:
                groups = OrderedDict(
                    (
                        k,
                        np.array([ g[k] for g in groups ], dtype=str)
                    ) for k in groups[0]
                )

            for k, v in groups.items():
                r = res[k]
                if k in block_handlers:
                    handler = block_handlers[k]
                else:
                    # unclear what we do in this case?
                    # Unless it breaks too much stuff, we'll throw an error...
                    # Originally, by default we just chose to not bother to handle it
                    # handler = None
                    raise StringParserException("{}: can't find a handler for group '{}' in block handlers ({})".format(
                        type(self).__name__,
                        k,
                        block_handlers
                    ))

                try:
                    self._handle_insert_result(r, handler, v, single=single, append=append)
                except Exception as e:
                    if append or (isinstance(append, int) and append == 0):
                        raise StringParserException(
                            "failed to append to key '{}' for results array {}; block handler={}, value={}".format(
                                k,
                                res,
                                handler,
                                v[:50] + "...<{}>...".format(len(v) - 50) + v[-50:] if len(v) > 100 else v
                            )) from e
                    else:
                        raise StringParserException("failed to insert into key '{}' for results array {}; block handler={}, value={}".format(
                            k,
                            res,
                            handler,
                            v[:50] + "...<{}>...".format(len(v) - 50) + v[-50:] if len(v) > 100 else v
                        )) from e


    @classmethod
    def _get_regex_handler(cls, r):
        # if we have a simple dtype we just parse out the whole thing, no need for parse_all
        # otherwise if we have non-None repetitions we need parse_all

        if r.handler is not None:
            return r.handler

        if r.dtype.is_simple and r.dtype.shape is None:
            handler = None
        else:
            # if not r.dtype.is_simple:

            # we should only ever end up here on Named and Capturing objects
            # these will cause infinite recursion if we don't strip their Capturing status somehow
            import copy
            r = copy.copy(r)
            r.capturing = False
            r.key = None
            r.invalidate_cache()

            # kids = r.named_groups
            # if kids is None:
            #     kids = r.capturing_groups
            # else:
            #     kids = list(kids.values())
            #     if kids[0] is r:
            #         kids = r.capturing_groups
            #
            # if len(kids) == 0:
            #     raise StringParserException("Can't create a block handler if there won't be a match...?")
            # elif len(kids) == 1:
            #     r = kids[0]

            def handler(t, r=r, cls = cls, array = None, append = False):
                """A default handler for a regex pattern, constructed to be flexible and to recurse cleanly
                If you lots of data, you might want to hook in your own handler

                :param t: text to parse
                :type t: str | iterable[str] | iterable[iterable[str]]
                :param r: the regex pattern we're subparsing with with
                :type r: RegexPattern
                :param cls: the StringParser class this comes from
                :type cls: type
                :param array: the results array passed through for shape determination and stuff
                :type array: StructuredTypeArray
                :return:
                :rtype:
                """

                if array is not None and array.can_cast(t):
                    rr = t
                elif isinstance(t, str):
                    rr = None
                    if len(t) > 0:
                        s = cls(r)
                        out_array = {'array':array, 'append':append}
                        s.parse(t, out = out_array)
                else:
                    # we assume we got a list of strings...?
                    rr = None
                    skipped_rows = 0 # needed to handle Optional patterns...
                    s = cls(r)
                    # at some point this will be pushed into a parse_vector
                    # function or something where we're more efficient
                    # about how we do the parsing and data caching

                    if array is not None:
                        # One thing that needs to happen here is that the code needs some indicator that the
                        # append depth level of the array needs to be incremented
                        # Unfortunately we also need some way to manage this such that the changes won't lead to
                        # depth blow-up
                        # Like I think it'd be good for people to be able to specify the append level to the array
                        # but we don't want the incremented append level to propagate back to here somehow...
                        array.append_depth += 1
                        append_depth = array.append_depth
                        if append is True or append is False:
                            append = 0
                        out_array = {
                            'array':array,
                            'append':append
                        }
                        # print(append_depth, append + append_depth, array)
                        app_depth = append + append_depth
                        array.set_filling(0, axis = app_depth)
                    # print("????", t)
                    for i, l in enumerate(t):
                        # we gotta basically go line-by-line and add our stuff in
                        # the way we do this is by applying parse to each one of these things
                        # then calling append to add all the data in
                        if array is not None:
                            if len(l) == 0:
                                array.append(None, axis = append)
                            else:
                                # print(l)
                                s.parse(l, dtypes=array.stype, out = out_array)
                            array.set_filling(i+1, axis = app_depth)
                            # if array.is_simple:
                            #     array.filled_to[app_depth] = i+1
                            # else:
                            #     print(l)

                        elif rr is None:

                            if len(l) == 0:
                                # nothing to do here, basically a 'None' return
                                skipped_rows += 1
                                continue

                            rr = s.parse(l)
                            if skipped_rows > 0:
                                def add_skips(r, skipped = skipped_rows):
                                    a = r.array
                                    if isinstance(a, np.ndarray):
                                        r.extend(a[:skipped], prepend = True)
                                    else:
                                        for a2 in a:
                                            add_skips(a2)
                            else:
                                add_skips = lambda p:None

                            # now that we have this object we need to basically add an axis to it
                            # the raw thing will be lower-dimensional that we want I think
                            # the way we handle that is then...? I guess hitting it all with an add_axis?
                            if isinstance(rr, StructuredTypeArray):
                                rr.add_axis()
                                add_skips(rr)
                            elif isinstance(rr, OrderedDict):
                                for k in rr:
                                    rr[k].add_axis()
                                    add_skips(rr[k])
                            else:
                                for r in rr:
                                    r.add_axis()
                                    add_skips(r)

                        else:
                            if len(l) == 0:
                                r2 = None
                            else:
                                r2 = s.parse(l)
                            # we might have an OrderedDict or we might have a list or we might have a StructuredArray
                            # gotta make sure all of these cases are clean
                            if isinstance(rr, StructuredTypeArray):
                                # print("y tho :>>", rr)
                                # print(r2)
                                rr.append(r2)
                                # print(rr)
                                # print("<<: y tho")
                            elif isinstance(rr, OrderedDict):
                                for k in rr:
                                    rr[k].append(r2[k])
                            else:
                                for r1, r2 in zip(rr, r2):
                                    r1.append(r2) # should I be using extend...?

                    # gotta do some shape management because otherwise the array never knows how many times we iterated
                    # through here...
                    if array is not None:
                        if append is True or append is False:
                            pass
                        else:
                            array.append_depth -= 1
                            app_depth = append + append_depth
                            if app_depth > 0:
                                array.increment_filling(axis=app_depth - 1)

                return rr

        return handler

    #endregion

    #region HandlerMethods
    @classmethod
    def handler_method(cls, method):
        """Turns a regular function into a handler method by adding in (and ignoring) the array argument

        :param method:
        :type method:
        :return:
        :rtype:
        """

        def handler(data, method = method, array = None, append = False):
            return method(data)

        return handler

    @staticmethod
    def load_array(data, dtype='float'):
        import io
        return np.loadtxt(io.StringIO(data), dtype=dtype)
    @classmethod
    def to_array(cls,
                 data,
                 array = None,
                 append = False,
                 dtype = 'float',
                 shape = None,
                 pre=None
                 ):
        """A method to take a string or iterable of strings and quickly dump it to a NumPy array of the right dtype (if it can be cast as one)

        :param data:
        :type data:
        :param dtype:
        :type dtype:
        :return:
        :rtype:
        """
        import io

        if not isinstance(data, str):
            if isinstance(data, np.ndarray) and data.ndim > 1:
                # we need to load each chunk in turn...
                def hold_me_closer_tiny_dancer(data, dtype = dtype, cls = cls):
                    if data.ndim == 1:
                        bleh = '\n'.join((d.strip() for d in data))
                        if pre is not None:
                            bleh = pre(bleh)
                        return cls.load_array(bleh, dtype = dtype)
                    else:
                        return [ hold_me_closer_tiny_dancer(d) for d in data ]
                arr = np.array(hold_me_closer_tiny_dancer(data))
            else:
                bleh = '\n'.join((d.strip() for d in data))
                if pre is not None:
                    bleh = pre(bleh)
                arr = cls.load_array(bleh, dtype=dtype)
        else:
            if pre is not None:
                data = pre(data)
            arr = cls.load_array(data, dtype=dtype)
        if shape is not None:
            fudge_axis = [ i for i, s in enumerate(shape) if s is None]
            if len(fudge_axis) == 0:
                arr = arr.reshape(shape)
            elif len(fudge_axis) == 1:
                fudge_axis = fudge_axis[0]
                num_els = np.prod(arr.shape)
                shoop = list(shape)
                shoop[fudge_axis] = 1
                num_not_on_axis = np.prod(shoop)
                if num_els % num_not_on_axis == 0:
                    num_left = num_els // num_not_on_axis
                    shoop[fudge_axis] = num_left
                    arr = arr.reshape(shoop)
                else:
                    raise StringParserException("{}.{}: can't reshape array with shape '{}' to shape '{}' as number of elements is not divisible by block size".format(
                        cls.__name__,
                        'to_array',
                        arr.shape,
                        shape
                    ))
            else:
                raise StringParserException("{}.{}: can't reshape to shape '{}' with more than one indeterminate slot".format(
                    cls.__name__,
                    'to_array',
                    shape
                ))

        return arr
    @classmethod
    def array_handler(cls,
                      array = None,
                      append = False,
                      dtype = 'float',
                      shape = None,
                      pre = None
                      ):
        """Returns a handler that uses to_array

        :param dtype:
        :type dtype:
        :param array:
        :type array:
        :param shape:
        :type shape:
        :return:
        :rtype:
        """
        def handler(data, array = array, append = append, dtype = dtype, shape = shape, cls=cls):
            return cls.to_array(data, array = array, append = append, dtype = dtype, shape = shape, pre=pre)
        return handler
    #endregion




from .HTML import HTML, CSS, ContentXML, HTMLManager
from .Bootstrap import Bootstrap
from .HTMLWidgets import ActiveHTMLWrapper, HTMLWidgets
from .BootstrapWidgets import BootstrapWidgets
from .WidgetTools import JupyterAPIs, DefaultOutputWidget

import functools

__all__ = [ "JHTML" ]
__reload_hook__ = [".HTML", ".HTMLWidgets", ".Bootstrap", ".BootstrapWidgets", ".WidgetTools"]


class JHTML:
    HTML = HTML
    HTMLManager = HTMLManager
    CSS = CSS
    XML = ContentXML
    HTMLWidgets = HTMLWidgets


    APIs = JupyterAPIs
    DefaultOutputWidget = DefaultOutputWidget

    """
    Provides dispatchers to either pure HTML components or Widget components based on whether interactivity
    is required or not
    """
    manage_class = HTMLManager.manage_class
    manage_style = HTMLManager.manage_styles
    extract_styles = HTMLManager.extract_styles
    manage_attrs = HTMLManager.manage_attrs
    @classmethod
    def load(cls, exec_prefix=None, overwrite=False):
        from IPython.core.display import display

        elems = [
            HTMLWidgets.load(exec_prefix=exec_prefix, overwrite=overwrite)
            # BootstrapWidgets.load()
        ]
        display(*(e for e in elems if e is not None))

    # RawHTML = JupyterAPIs.raw_html
    @classmethod
    def Markdown(cls, text):
        import markdown
        html_string = markdown.markdown(text, extensions=['fenced_code'])
        # print(html_string)
        return cls.parse(html_string, strict=False, strip=False)

    def __init__(self, context=None,
                 include_bootstrap=False,
                 expose_classes=False,
                 output_pane=True,
                 callbacks=None,
                 widgets=None
                 ):
        self._context = context
        self._additions = set()
        self._include_boostrap = include_bootstrap
        self.expose_classes = expose_classes
        self._callbacks = callbacks
        self._widgets = widgets
        self._output_pane = (
            DefaultOutputWidget() if output_pane is True
            else DefaultOutputWidget(output_pane) if output_pane is not None
            else output_pane
        )

    def _get_frame_vars(self):
        import inspect
        frame = inspect.currentframe()
        parent = frame.f_back.f_back.f_back
        # print(parent.f_locals)
        return parent.f_locals
    def insert_vars(self):
        globs = self._context
        if globs is None:
            globs = self._get_frame_vars()
        cls = type(self)
        for x in cls.__dict__.keys():
            if not x.startswith("_"):
                globs[x] = getattr(cls, x)
                self._additions.add(x)
        if self._include_boostrap:
            for x in cls.Bootstrap.__dict__.keys():
                if not x.startswith("_"):
                    globs[x] = getattr(cls.Bootstrap, x)
                    self._additions.add(x)

    def wrap_callbacks(self, c):
        if self._output_pane is None:
            callbacks = c
        else:
            callbacks = {}
            for k,v in c.items():
                @functools.wraps(v)
                def callback(*args, self=self, **kwargs):
                    with self._output_pane:
                        v(*args, **kwargs)
                callbacks[k] = callback
        return callbacks

    _callback_stack = []
    _widget_stack = []
    def __enter__(self):
        """
        To make writing HTML interactively a bit nicer

        :return:
        :rtype:
        """

        if self.expose_classes:
            self.insert_vars()
        if self._callbacks is not None:
            cls = type(self)
            self._callback_stack.append(cls.callbacks.copy())
            cls.callbacks = self.wrap_callbacks(self._callbacks)
        if self._output_pane is not None:
            self._output_pane.__enter__()
            if self._widgets is None:
                self._widgets = {'out':self._output_pane.obj}
            elif 'out' not in self._widgets:
                self._widgets['out'] = self._output_pane.obj
            else:
                self._widgets['default_output'] = self._output_pane.obj
        if self._widgets is not None:
            cls = type(self)
            self._widget_stack.append(cls.widgets.copy())
            cls.widgets = self._widgets

        return self
    @property
    def out(self):
        return self._output_pane.obj

    def prune_vars(self):
        globs = self._context
        if globs is None:
            globs = self._get_frame_vars()
        for x in self._additions:
            try:
                del globs[x]
            except KeyError:
                pass
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.expose_classes:
            self.prune_vars()
        if self._output_pane is not None:
            self._output_pane.__exit__(exc_type, exc_val, exc_tb)
        if self._callbacks is not None:
            cls = type(self)
            cls.callbacks = self._callback_stack.pop()
        if self._widgets is not None:
            cls = type(self)
            cls.widgets = self._widget_stack.pop()

    callbacks = {}
    @classmethod
    def parse_handlers(cls, handler_string):
        handlers = {}
        for x in handler_string.split(";"):
            k,v = x.split(":")
            k = k.strip()
            v = v.strip()
            if v in cls.callbacks:
                v = cls.callbacks[v]
            if isinstance(v, str):
                v = globals()[v]
            handlers[k] = v

        return handlers

    widgets = {} # should this be a weakset...?
    @classmethod
    def parse_widget(cls, uuid):
        if uuid in cls.widgets:
            return cls.widgets[uuid]
        else:        
            widgets = JupyterAPIs.get_widgets_api().Widget.widgets.values()
            map = {w.model_id: w for w in widgets}
            return map[uuid]
    @classmethod
    def convert(cls, etree, strip=True, converter=None, **extra_attrs):
        if converter is None:
            converter = cls.convert

        base = HTML.convert(etree, strip=strip, converter=converter, **extra_attrs)
        if base.tag.lower() == "jwidget":
            base = cls.parse_widget(base.attrs['id'])
        else:
            attrs = base.attrs
            handlers = cls.parse_handlers(attrs['event-handlers']) if 'event-handlers' in attrs else None
            track_value = attrs['track-value'].lower() == 'true' if 'track-value' in attrs else False
            dynamic = attrs['dynamic'].lower() == 'true' if 'dynamic' in attrs else False
            for k in ['event-handlers', 'track-value', 'dynamic']:
                try:
                    del attrs[k]
                except KeyError:
                    ...
            base.attrs = attrs
            if not dynamic:
                dynamic = track_value or handlers is not None
            if not dynamic:
                Widget = JupyterAPIs.get_widgets_api().Widget
                dynamic = any(isinstance(x, (ActiveHTMLWrapper, Widget)) for x in base.elems)
            if dynamic:
                base = HTMLWidgets.from_HTML(base, track_value=track_value, event_handlers=handlers)

        return base

    @classmethod
    def parse(cls, src, event_handlers=None, dynamic=None, track_value=None, strict=True, fallback=None, **attrs):
        base = HTML.parse(src, strict=strict, fallback=fallback, converter=cls.convert)
        _debugPrint = False if '_debugPrint' not in attrs else attrs['_debugPrint']
        trackInput = False if 'trackInput' not in attrs else attrs['trackInput']
        if isinstance(base, HTML.XMLElement) and (
                event_handlers is not None
                or track_value is True
                or trackInput is True
                or _debugPrint is True
                or dynamic is True
        ):
            base = HTMLWidgets.from_HTML(base,
                                        event_handlers=event_handlers,
                                        track_value=track_value,
                                        _debugPrint=_debugPrint,
                                        **attrs
                                        )

        return base

    @classmethod
    def _check_widg(cls, elems):
        wapi = JupyterAPIs.get_widgets_api()
        if wapi is None:
            return False

        Widget = JupyterAPIs.get_widgets_api().Widget
        if isinstance(elems, (list, tuple)) and len(elems) == 0:
            return False
        elif isinstance(elems, (ActiveHTMLWrapper, Widget)) or hasattr(elems, 'to_widget'):
            return True
        else:
            return any(
                cls._check_widg(e) if isinstance(e, (list, tuple)) else
                 (isinstance(e, (ActiveHTMLWrapper, Widget)) or hasattr(e, 'to_widget'))
                for e in elems
            )

    @classmethod
    def _resolve_source(jhtml, plain, widget, *elems,
                        event_handlers=None, dynamic=None,
                        track_value=None, trackInput=None, _debugPrint=None,
                        javascript_handles=None, oninitialize=None,
                        **attrs):
        if (
            event_handlers is not None
            or track_value is True
            or trackInput is True
            or _debugPrint is True
            or dynamic is True
            or javascript_handles is not None
            or oninitialize is not None
        ):
            return widget
        else:
            if jhtml._check_widg(elems):
                return widget
            else:
                return plain
    @classmethod
    def _dispatch(jhtml, plain, widget, *elements, event_handlers=None, dynamic=None, **styles):
        src = jhtml._resolve_source(plain, widget, *elements, event_handlers=event_handlers, dynamic=dynamic, **styles)
        if src is plain:
            return src(*elements, activator=widget.from_HTML, **styles)
        else:
            return src(*elements, event_handlers=event_handlers, **styles)

    def dispatcher(fn):
        name = fn.__name__
        @functools.wraps(fn)
        def dispatcher(jhtml, *elements, **attrs):
            return jhtml._dispatch(getattr(HTML, name), getattr(HTMLWidgets, name), *elements, **attrs)
        return dispatcher

    @classmethod
    @dispatcher
    def Abbr(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Address(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Anchor(jhtml, *elements, **styles): ...
    A = Anchor
    @classmethod
    @dispatcher
    def Area(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Article(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Aside(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Audio(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def B(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Base(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Bdi(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Bdo(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Blockquote(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Body(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Bold(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Br(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Button(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Canvas(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Caption(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Cite(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Code(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Col(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Colgroup(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Data(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Datalist(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Dd(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Del(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Details(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Dfn(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Dialog(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Div(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Dl(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Dt(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Em(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Embed(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Fieldset(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Figcaption(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Figure(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Footer(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Form(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Head(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Header(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Heading(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Hr(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Html(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Iframe(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Image(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Img(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Input(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Ins(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Italic(jhtml, *elements, **styles): ...
    I = Italic
    @classmethod
    @dispatcher
    def Kbd(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Label(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Legend(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Link(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def List(jhtml, *elements, **styles): ...
    Ul = List
    @classmethod
    @dispatcher
    def ListItem(jhtml, *elements, **styles): ...
    Li = ListItem
    @classmethod
    @dispatcher
    def Main(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Map(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Mark(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Meta(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Meter(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Nav(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Noscript(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def NumberedList(jhtml, *elements, **styles): ...
    Ol = NumberedList
    @classmethod
    @dispatcher
    def Object(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Optgroup(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Option(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Output(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Param(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Picture(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Pre(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Progress(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Q(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Rp(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Rt(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Ruby(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def S(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Samp(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Script(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Section(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Select(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Small(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Source(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Span(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Strong(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Style(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Sub(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def SubHeading(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def SubsubHeading(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def SubsubsubHeading(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def SubHeading5(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def SubHeading6(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Summary(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Sup(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Svg(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Table(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def TableBody(jhtml, *elements, **styles): ...
    Tbody = TableBody
    @classmethod
    @dispatcher
    def TableFooter(jhtml, *elements, **styles): ...
    Tfoot = TableFooter
    @classmethod
    @dispatcher
    def TableHeader(jhtml, *elements, **styles): ...
    Thead = TableHeader
    @classmethod
    @dispatcher
    def TableHeading(jhtml, *elements, **styles): ...
    Th = TableHeading
    @classmethod
    @dispatcher
    def TableItem(jhtml, *elements, **styles): ...
    Td = TableItem
    @classmethod
    @dispatcher
    def TableRow(jhtml, *elements, **styles): ...
    Tr = TableRow
    @classmethod
    @dispatcher
    def Template(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Text(jhtml, *elements, **styles): ...
    P = Text
    @classmethod
    @dispatcher
    def Textarea(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Time(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Title(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Track(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def U(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Var(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Video(jhtml, *elements, **styles): ...
    @classmethod
    @dispatcher
    def Wbr(jhtml, *elements, **styles): ...

    del dispatcher

    OutputArea = HTMLWidgets.OutputArea
    JavascriptAPI = HTMLWidgets.JavascriptAPI

    class Bootstrap:

        Class = Bootstrap.Class
        Variant = Bootstrap.Variant

        @classmethod
        def _dispatch(jhtml, *elems, **attrs):
            return JHTML._dispatch(*elems, **attrs)
        def dispatcher(fn):
            name = fn.__name__
            @functools.wraps(fn)
            def dispatcher(boots, *elements, **attrs):
                return boots._dispatch(getattr(Bootstrap, name), getattr(BootstrapWidgets, name), *elements, **attrs)
            return dispatcher

        @classmethod
        @dispatcher
        def Icon(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Alert(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Badge(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def PanelBody(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def PanelHeader(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Panel(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def CardBody(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def CardHeader(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def CardFooter(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def CardImage(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Card(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Jumbotron(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Col(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Row(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Container(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Button(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def ButtonGroup(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def LinkButton(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Table(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def ListGroup(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def ListGroupItem(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def FontAwesomeIcon(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def GlyphIcon(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Label(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Pill(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def ListComponent(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def ListItemComponent(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def Breadcrumb(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def BreadcrumbItem(boots, *elements, **styles): ...


        @classmethod
        @dispatcher
        def Accordion(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def AccordionItem(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def AccordionCollapse(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def AccordionBody(boots, *elements, **styles): ...

        @classmethod
        @dispatcher
        def Carousel(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def CarouselInner(boots, *elements, **styles): ...
        @classmethod
        @dispatcher
        def CarouselItem(boots, *elements, **styles): ...

        @classmethod
        @dispatcher
        def Collapse(boots, *elements, **styles): ...

        del dispatcher

    class Styled:
        def __init__(self, base, **attrs):
            self.attrs = attrs
            self.base = base
        def __call__(self, *args, **kwargs):
            return self.base(*args, **dict(self.attrs, **kwargs))
        def __repr__(self):
            return "{}({}, {})".format(
                type(self).__name__,
                self.base,
                self.attrs
            )
    class Compound:
        def __init__(self, *wrappers):
            self.base = wrappers[-1]
            self.classes = wrappers[:-1]
        @staticmethod
        def destructure_wrapper(wrapper):
            try:
                lw = len(wrapper)
            except TypeError:
                lw = 0
            if lw == 2 and isinstance(wrapper[0], str):
                name, wrapper = wrapper
            else:
                name = None
            return name, wrapper
        class CompoundWrapperData:
            __slots__ = ["list", "dict"]
            def __init__(self, list, dict):
                self.list = list
                self.dict = dict
            def __getitem__(self, item):
                if isinstance(item, str):
                    return self.dict[item]
                else:
                    return self.list[item]
        def __call__(self, *args, wrapper_attrs=None, **kwargs):
            cache = []
            names = {}
            if wrapper_attrs is None:
                wrapper_attrs = {}
            cwd = self.CompoundWrapperData(cache, names)
            n,w = self.destructure_wrapper(self.base)
            base = w(*args, **kwargs, **wrapper_attrs.get(n, {}))
            if n is not None:
                names[n] = base
            cache.append(base)
            base.compound_wrapper_data = cwd
            for c in reversed(self.classes):
                n, c = self.destructure_wrapper(c)
                base = c(base, **wrapper_attrs.get(n, {}))
                if n is not None:
                    names[n] = base
                cache.append(base)
                base.compound_wrapper_data = cwd
            return base
        def __repr__(self):
            return "{name}({spacing}{objs}{spacing})".format(
                name=type(self).__name__,
                objs=",\n    ".join(repr(x) for x in self.classes+(self.base,)),
                spacing="\n    " if len(self.classes) > 0 else ""
            )


import uuid

import numpy as np
from .CommonCoordinateSystems import InternalCoordinateSystem, CartesianCoordinates3D, ZMatrixCoordinateSystem, ZMatrixCoordinates
from .CoordinateSystemConverter import CoordinateSystemConverter
from .ZMatrixToCartesian import ZMatrixToCartesianConverter
from .CartesianToZMatrix import CartesianToZMatrixConverter
from ... import Numputils as nput
# import weakref

__all__ = [
    "IterativeZMatrixCoordinateSystem",
    "IterativeZMatrixCoordinates",
    "CartesianToIZSystemConverter",
    "IZSystemToCartesianConverter"
]

#TODO: these should all be metaclasses but :shrag:
class IterativeZMatrixCoordinateSystem(ZMatrixCoordinateSystem):
    """
    Represents ZMatrix coordinates generally
    """
    name = "IterativeZMatrix"

    def __init__(self,
                 converter_options=None,
                 dimension=(None, None),
                 coordinate_shape=(None, 3),
                 **opts):
        """
        :param converter_options: options to be passed through to a `CoordinateSystemConverter`
        :type converter_options: None | dict
        :param coordinate_shape: shape of a single coordinate in this coordiante system
        :type coordinate_shape: Iterable[None | int]
        :param dimension: the dimension of the coordinate system
        :type dimension: Iterable[None | int]
        :param opts: other options, if `converter_options` is None, these are used as the `converter_options`
        :type opts:
        """
        super().__init__(dimension=dimension, coordinate_shape=coordinate_shape, converter_options=converter_options, **opts)
IterativeZMatrixCoordinates = IterativeZMatrixCoordinateSystem()
IterativeZMatrixCoordinates.__name__ = "IterativeZMatrixCoordinates"
IterativeZMatrixCoordinates.__doc__ = """Iterative Z-matrix internals"""

class CartesianToIZSystemConverter(CartesianToZMatrixConverter):
    """
    A converter class for going from Cartesian coordinates to internals coordinates
    """

    @property
    def types(self):
        return (CartesianCoordinates3D, IterativeZMatrixCoordinates)

    def convert_many(self, coords, *, ordering, use_rad=True, return_derivs=False, **kw):
        ints, opts = super().convert_many(
            coords=coords,
            ordering=ordering,
            use_rad=use_rad,
            return_derivs=return_derivs,
            **kw
        )
        opts['reference_coordinates'] = coords
        # opts['reference_internals'] = ints
        return ints, opts

class IZSystemToCartesianConverter(CoordinateSystemConverter):
    """
    A converter class for going from Cartesian coordinates to internals coordinates
    """

    @property
    def types(self):
        return (IterativeZMatrixCoordinates, CartesianCoordinates3D)

    def convert_many(self, coords, *, reference_coordinates, order=0,
                     masses=None,
                     remove_translation_rotation=True,
                     derivs=None,
                     return_derivs=None,
                     ordering=None,
                     origins=None,
                     axes=None,
                     embedding_coords=None,
                     jacobian_prep=None,
                     axes_labels=None,
                     fixed_atoms=None,
                     use_rad=True,
                     **kw):
        """
        We'll implement this by having the ordering arg wrap around in coords?
        """

        if return_derivs is None or return_derivs is True:
            return_derivs = order
        if return_derivs == 0: return_derivs = 1
        if not nput.is_numeric(return_derivs):
            return_derivs = max(return_derivs)

        sub_embedding = [1, 2, 5]
        def conv(carts, order=None, sys=coords.system):
            from .CoordinateSet import CoordinateSet
            base_shape = carts.shape[:-2]
            carts = CoordinateSet(carts, CartesianCoordinates3D)
            ints = carts.convert(sys,
                                 origins=origins, axes=axes, ordering=ordering,
                                 # embedding_coords=embedding_coords
                                 use_rad=use_rad
                                 )
            derivs = carts.jacobian(sys,
                                    order=order,
                                    converter_options=dict(
                                        origins=origins, axes=axes, ordering=ordering,
                                        use_rad=use_rad
                                        # embedding_coords=embedding_coords
                                    ),
                                    jacobian_prep=jacobian_prep,
                                    axes_labels=axes_labels
                                    )
            derivs = [
                d.reshape(base_shape + (carts.shape[-2] * 3,) * (n + 1) + (d.shape[-2] * d.shape[-1],))
                for n, d in enumerate(derivs)
            ]

            ints = ints.reshape(base_shape + (-1,)).view(np.ndarray)
            ints = np.delete(ints, sub_embedding, axis=-1)

            derivs = [
                np.delete(d, sub_embedding, axis=-1)
                for d in derivs
            ]

            return [ints] + derivs
        flat_coords = coords.reshape(coords.shape[:-2] + (-1,)).view(np.ndarray)
        flat_coords = np.delete(flat_coords, sub_embedding, axis=-1)
        (expansions, errors), _ = nput.inverse_coordinate_solve(conv,
                                                                flat_coords,
                                                                reference_coordinates,
                                                                order=return_derivs,
                                                                return_expansions=True,
                                                                return_internals=True,
                                                                masses=masses,
                                                                remove_translation_rotation=False,
                                                                fixed_atoms=fixed_atoms,
                                                                **kw
                                                                )
        carts, derivs = expansions[0], expansions[1:]
        opts = {
            # 'specs':specs,
            'derivs': derivs,
            'remove_translation_rotation': remove_translation_rotation
        }
        if masses is not None:
            opts['masses'] = masses

        return carts, opts

    def convert(self, coords, *, reference_coordinates, specs, order=0, **kw):
        return self.convert_many(coords, reference_coordinates=reference_coordinates, specs=specs, order=order, **kw)

__converters__ = [CartesianToIZSystemConverter(), IZSystemToCartesianConverter()]


"""
LazyTensors provides a small framework for symbolically working with Tensors
"""

import numpy as np
from ..Numputils import SparseArray

__all__ = [
    'Tensor',
    'TensorOp',
    'LazyOperatorTensor',
    'SparseTensor'
]

########################################################################################################################
#
#                                           Tensor
#
class Tensor:
    """A semi-symbolic representation of a tensor. Allows for lazy processing of tensor operations."""
    def __init__(self, a, shape = None):
        self._a = a
        self._shape = shape
    @classmethod
    def from_array(cls, a, shape = None):
        if isinstance(a, Tensor):
            return a
        else:
            return cls(a, shape = shape)
    @property
    def array(self):
        if isinstance(self._a, np.ndarray):
            return self._a
        else:
            return self._a.array

    def get_shape(self, a):
        return a.shape
    @property
    def shape(self):
        return self.get_shape(self._a)
    def get_dim(self):
        return len(self.shape)
    @property
    def dim(self):
        return self.get_dim()

    def add(self, other, **kw):
        return TensorPlus(self, other, **kw)
    def mul(self, other, **kw):
        return TensorMul(self, other, **kw)
    def dot(self, other, **kw):
        return TensorDot(self, other, **kw)
    def transpose(self, axes, **kw):
        return TensorTranspose(self, axes, **kw)
    def pow(self, other, **kw):
        return TensorPow(self, other, **kw)

    def __mul__(self, other):
        if isinstance(other, float):
            return TensorMul(self, other)
        else:
            return TensorDot(self, other)
    def __rmul__(self, other):
        if isinstance(other, float):
            return TensorMul(self, other)
        else:
            return TensorDot(self, other)
    def __add__(self, other):
        return TensorPlus(self, other)
    def __pow__(self, power, modulo=None):
        return TensorPow(self, power)

    def handle_missing_indices(self, missing, extant):
        # the default assumption is basically to use the diagonal element -- i.e. to assume that only the extant indices matter
        return self._a[extant]
    def pull_index(self, *idx):
        """Defines custom logic for handling how we pull indices

        :param idx:
        :type idx:
        :return:
        :rtype:
        """
        dim = len(idx)
        sdim = self.dim
        if dim > sdim:
            raise ValueError("{} has only {} dimensions but {} were requested".format(self, sdim, dim))
        else:
            real_shape = self._a.shape
            # here we add our hacks
            real_dim = len(real_shape)
            if real_dim < dim:
                a = self.handle_missing_indices(idx[:dim-real_dim], idx[dim-real_dim:])
            else:
                a = self._a[idx]
            try:
                new = type(self)(a)
            except ValueError:
                new = Tensor(a)
            return new

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.pull_index(item)
        else:
            return self.pull_index(*item)

    def __repr__(self):
        return "{}(<{}>)".format(type(self).__name__, ", ".join((str(x) for x in self.shape)))

class SparseTensor(Tensor):
    """
    Tensor class that uses SparseArray
    """

    def __init__(self, a, shape = None):
        super().__init__(SparseArray(a), shape = shape)

    @property
    def array(self):
        return self._a.todense()

########################################################################################################################
#
#                                           TensorOps
#
class TensorOp(Tensor):
    """A lazy representation of tensor operations to save memory"""
    def __init__(self, a, b, axis = None):
        # if isinstance(a, Tensor):
        #     a = a._a
        # if isinstance(b, Tensor):
        #     a = b._a
        self._a = a
        self._b = b
        if axis is None:
            self._axis = None
            self._axis_2 = None
        elif isinstance(axis, int):
            self._axis = axis
            self._axis_2 = None
        else:
            self._axis = axis[1]
            self._axis_2 = axis[0]
        self._kw = dict(axis = axis)

    def op(self, a, b):
        raise NotImplementedError(
            "{}: {} not implemented until a subclass does it".format(type(self.__name__), 'op')
        )
    def get_shape(self, a, b):
        return a.shape
    @property
    def shape(self):
        return self.get_shape(self._a, self._b)
    @property
    def array(self):
        """Ought to always compile down to a proper ndarray

        :return:
        :rtype: np.ndarray
        """
        return self.op(self._a, self._b)
    def __getitem__(self, i):
        return self.op(self._a[i], self._b)
class TensorPlus(TensorOp):
    """Represents an addition of two tensors"""
    def op(self, a, b):
        if isinstance(a, Tensor):
            a = a.array # with addition we can't really avoid it...
        if isinstance(b, Tensor):
            b = b.array
        return a + b
    def get_shape(self, a, b):
        return super().get_shape(a, b)
    def __getitem__(self, i):
        return type(self)(self._a[i], self._b[i])
class TensorPow(TensorOp):
    """Represents a power of a tensors"""
    def op(self, a, b):
        if isinstance(a, Tensor):
            a = a.array
        return np.linalg.matrix_power(a, b)
    def get_shape(self, a, b):
        return super().get_shape(a, b)
    def __getitem__(self, i):
        return type(self)(self._a[i], self._b)
class TensorMul(TensorOp):
    """Represents a multiplication of a tensor and a scalar"""
    def op(self, a, b):
        if isinstance(a, Tensor):
            a = a.array # with scalar mult we can't really avoid it...
        return a * b
    def get_shape(self, a, b):
        return super().get_shape(a, b)
    def __getitem__(self, i):
        return type(self)(self._a[i], self._b)
class TensorTranspose(TensorOp):
    """Represents a tensor transposition"""
    def get_shape(self, a, b):
        a_shp = a.shape
        return tuple(a_shp[a] for a in b)
    def op(self, a, b):
        if isinstance(a, Tensor):
            a = a.array
        return a.transpose(b)
class TensorDot(TensorOp):
    """Represents a tensor contraction"""
    def get_shape(self, a, b):
        a_shp = a.shape
        b_shp = b.shape
        ax = 0 if self._axis is None else self._axis
        ax2 = -1 if self._axis_2 is None else self._axis_2
        ax = set(len(b_shp) + a if a < 0 else a for a in ((ax,) if isinstance(ax, int) else ax))
        ax2 = set(len(a_shp) + a if a < 0 else a for a in ((ax2,) if isinstance(ax2, int) else ax2))
        shared = 0
        for shared, s in enumerate(zip(a_shp, b_shp)):
            if s[0] != s[1]:
                break
        shared = min(shared, min(ax), min(ax2))

        return (
            a_shp[:shared] +
            tuple(a for i,a in enumerate(a_shp[shared:]) if i+shared not in ax2) +
            tuple(a for i,a in enumerate(b_shp[shared:]) if i+shared not in ax)
        )
    def op(self, a, b):
        from ..Numputils import vec_tensordot

        # we'll allow this to cast to dense as we'll assume people are only ever calling this on low dimensional tensors...
        if isinstance(a, Tensor):
            a = a.array
        if isinstance(b, Tensor):
            b = b.array
        ax = 0 if self._axis is None else self._axis
        ax2 = -1 if self._axis_2 is None else self._axis_2
        if isinstance(ax, int):
            ax = (ax,)
        if isinstance(ax2, int):
            ax2 = (ax2,)
        contract = vec_tensordot(a, b, axes = (ax2, ax))
        return contract
    def __getitem__(self, i):
        return type(self)(self._a[i], self._b, **self._kw)

########################################################################################################################
#
#                                           LazyOperatorTensor
#
class LazyOperatorTensor(Tensor):
    """A super-lazy tensor that represents the elements of an operator """

    def __init__(self, operator, shape, memoization = True, dtype = object, fill = None):
        if memoization is True:
            memoization = {}
        self.memoization = isinstance(memoization, dict)
        if self.memoization:
            a = memoization
        else:
            a = None

        self.operator = operator
        self.dtype = dtype
        self.fill = fill
        super().__init__(a, shape=shape)

    @property
    def array(self):
        import itertools as it

        base = np.full(self.shape, self.fill, dtype=self.dtype)
        mem = self._a
        try:
            self._a = None
            _cast = False
            for idx in it.product(range(x) for x in self.shape):
                res = self._get_element(idx)
                if not _cast:
                    if isinstance(res, (int, float, np.integer, np.float)):
                        base = np.zeros(base, dtype=type(res))
                    _cast = True

                base[idx] = res
        finally:
            self._a = mem

        return base

    def _get_element(self, indices):
        if self._a is not None:
            try:
                res = self._a[indices]
            except (KeyError, IndexError):
                res = self.operator(indices)
                self._a[indices] = res
            return res
        else:
            return self.operator(indices)

    def __getitem__(self, item):
        return self._get_element(item)



import abc
import enum
import numpy as np
from .. import Devutils as dev
from .. import Numputils as nput

from .Characters import *
from .Elements import *

__all__ = [
    "NamedPointGroups",
    "ParametrizedPointGroups",
    "PointGroup"
]

class NamedPointGroups(enum.Enum):
    Ih = "Ih"
    I = "I"
    Oh = "Oh"
    O = "O"
    Th = "Th"
    Td = "Td"
    T = "T"
    Cs = "Cs"
class ParametrizedPointGroups(enum.Enum):
    C = "C"
    Ch = "Ch"
    Cv = "Cv"
    D = "D"
    Dh = "Dh"
    Dd = "Dd"
    S = "S"


class PointGroup(metaclass=abc.ABCMeta):
    group: 'NamedPointGroups|ParametrizedPointGroups'

    def __init__(self, character_table=None, elements=None, axes=None):
        self._elements = elements
        self._character_table = character_table
        self._axes = axes
        self._base_axes = None

    def get_modification_kwargs(
            self,
            character_table=dev.default,
            elements=dev.default,
            axes=dev.default
    ):
        return dict(
            character_table=self.character_table if dev.is_default(character_table) else character_table,
            elements=elements if dev.is_default(elements) else elements,
            axes=axes if dev.is_default(axes) else axes,
        )
    def modify(self,
               character_table=dev.default,
               elements=dev.default,
               axes=dev.default
               ):
        return type(self)(**self.get_modification_kwargs(
            character_table=character_table,
            elements=elements,
            axes=axes
        ))

    @abc.abstractmethod
    def get_name(self):
        ...
    @property
    def name(self):
        return self.get_name()
    def __repr__(self):
        return "PointGroup<{}>".format(self.name)

    @classmethod
    def from_name(cls, key, n=None, **etc):
        if isinstance(key, NamedPointGroups):
            return NamedPointGroup(key, **etc)
        elif isinstance(key, ParametrizedPointGroups):
            return ParametrizedPointGroup(key, n, **etc)
        else:
            if any(x.isdigit() for x in key):
                n = int("".join(x for x in key if x.isdigit()))
                key = "".join("".join(x for x in key if not x.isdigit()))
                return cls.from_name(key, n=n, **etc)
            try:
                key = NamedPointGroups(key)
            except ValueError:
                key = ParametrizedPointGroups(key)
            return cls.from_name(key, n=n, **etc)

    @classmethod
    def get_symmetry_element_primary_rotation(cls, elements: 'Iterable[SymmetryElement]'):
        rotors = [e for e in elements if isinstance(e, RotationElement)]
        if len(rotors) == 0:
            return None, None
        else:
            orders, loc, counts = np.unique([e.order for e in rotors], return_counts=True, return_index=True)
            return rotors[loc[-1]], counts[-1]

    @classmethod
    def from_symmetry_elements(cls, elements: list[SymmetryElement], tol=1e-2):
        primary_axis, counts = cls.get_symmetry_element_primary_rotation(elements)
        if primary_axis is None:
            if any(isinstance(e, InversionElement) for e in elements):
                return cls.from_name("Ci")
            elif any(isinstance(e, ReflectionElement) for e in elements):
                reflections = [e for e in elements if isinstance(e, ReflectionElement)]
                axes = nput.view_matrix(reflections[0].axis)[:, (0, 2, 1)]
                return cls.from_name("Cs", axes=axes)
            else:
                return cls.from_name("C", 1)
        elif counts > 1:  # high symmetry
            raise NotImplementedError(...)
        else:
            group_order = primary_axis.order
            primary_axis = primary_axis.axis
            c2_axis = None
            for e in elements:
                if (
                        isinstance(e, RotationElement)
                        and e.order == 2
                        and np.abs(np.dot(e.axis, primary_axis)) < tol  # perp
                ):
                    c2_axis = e
                    break

            sh_plane = None
            for e in elements:
                if (
                        isinstance(e, ReflectionElement)
                        and np.abs(np.dot(e.axis, primary_axis)) > 1 - tol  # parallel
                ):
                    sh_plane = e
                    break

            if c2_axis is not None:
                if sh_plane is not None:
                    return cls.from_name("Dh", group_order)
                else:
                    sd_plane = None
                    axes = nput.view_matrix(
                        primary_axis.axis,
                        view_vector=c2_axis.axis
                    )[:, (0, 2, 1)]
                    for e in elements:
                        if (
                                isinstance(e, ReflectionElement)
                                and np.abs(np.dot(e.axis, primary_axis)) < tol  # perp
                        ):
                            sd_plane = e
                    if sd_plane is not None:
                        return cls.from_name("Dd", group_order, axes=axes)
                    else:
                        return cls.from_name("D", group_order, axes=axes)
            else:
                if sh_plane is not None:
                    return cls.from_name("Ch", group_order)
                else:
                    sv_plane = None
                    for e in elements:
                        if (
                                isinstance(e, ReflectionElement)
                                and np.abs(np.dot(e.axis, primary_axis)) < tol  # perp
                        ):
                            sv_plane = e
                    if sv_plane is not None:
                        return cls.from_name("Cv", group_order)
                    else:
                        if any(isinstance(e, ImproperRotationElement) for e in elements):
                            return cls.from_name("S", group_order)
                        else:
                            return cls.from_name("C", group_order)

    def get_symmetry_elements(self, only_class_representatives=True):
        ct = self.character_table
        if only_class_representatives:
            mats = ct.matrices
            if mats is None:
                raise NotImplementedError(f"no implementation for operator matrices for point group {self}")
            elems = tuple(SymmetryElement.from_transformation_matrix(x) for x in mats)
            self._base_axes = self.get_axes(elems)
            if self._axes is not None:
                tf = self._axes @ self._base_axes.T
                if abs(np.linalg.det(tf)) < 1e-6:
                    raise ValueError("bad transformation")
                elems = [e.transform(tf) for e in elems]
        else:
            all_mats = self.get_all_character_matrices()
            elems = tuple(SymmetryElement.from_transformation_matrix(x) for x in all_mats)
            if self._axes is not None:
                mats = ct.matrices
                if mats is None:
                    raise NotImplementedError(f"no implementation for operator matrices for point group {self}")
                base_elems = tuple(SymmetryElement.from_transformation_matrix(x) for x in mats)
                base_axes = self._base_axes
                if base_axes is None:
                    base_axes = self.get_axes(base_elems)
                tf = self._axes @ base_axes.T
                elems = [e.transform(tf) for e in elems]

        return elems

    @property
    def character_table(self) -> CharacterTable:
        if self._character_table is None:
            self._character_table = self.get_character_table()
        return self._character_table

    @property
    def elements(self) -> 'tuple[SymmetryElement]':
        if self._elements is None:
            self._elements = self.get_symmetry_elements()
        return self._elements

    @abc.abstractmethod
    def get_character_table(self):
        ...

    @abc.abstractmethod
    def get_all_character_matrices(self):
        ...

    @classmethod
    def get_axes_from_symmetry_elements(cls, elements):
        primary_axis = None
        secondary_axis = None

        primary, count = cls.get_symmetry_element_primary_rotation(elements)
        if primary is None:
            # search for reflection plane
            planes = [e for e in elements if isinstance(e, ReflectionElement)]
            if len(planes) == 0:
                return np.eye(3)
            primary_axis = planes[0].axis
            if len(planes) > 1: # I don't think this is possible
                secondary_axis = planes[1].axis
        else:
            primary_axis = primary.axis
            c2_axes = [e for e in elements if isinstance(e, RotationElement) and e.order == 2]
            perp_axes = [
                e for e in c2_axes
                if abs(np.dot(primary_axis, e.axis)) < 1e-2
            ]
            if len(perp_axes) == 0:
                # search for vertical reflection planes
                planes = [e for e in elements if isinstance(e, ReflectionElement)]
                v_planes = [
                    e for e in planes
                    if abs(np.dot(primary_axis, e.axis)) < 1e-2
                ]
                if len(v_planes) > 0:
                    secondary_axis = v_planes[0].axis
                else:
                    non_par = [
                        e for e in c2_axes
                        if abs(np.dot(primary_axis, e.axis)) < .9
                    ]
                    if len(non_par) > 0:
                        secondary_axis = non_par[0].axis
            else:
                secondary_axis = perp_axes[0].axis

        if primary_axis is None:
            raise ValueError("no primary axis found") # how??

        if secondary_axis is None:
            secondary_axis = [1, 0, 0]
            if abs(np.dot(secondary_axis, primary_axis)) > 1 - 1e-2:
                secondary_axis = [0, 0, 1]

        return nput.view_matrix(primary_axis, secondary_axis, output_order=(0, 2, 1))

    def get_axes(self, elements=None, base_axes=False):
        # could use point group identity for short circuiting, but don't want to
        if base_axes:
            elements = tuple(
                SymmetryElement.from_transformation_matrix(x)
                for x in self.character_table.matrices
            )
        if elements is None:
            elements = self.elements
        return self.get_axes_from_symmetry_elements(elements)
        # tertiary_axis = nput.vec_crosses(primary_axis, secondary_axis, normalize=True)
        # secondary_axis = nput.vec_crosses(tertiary_axis, primary_axis, normalize=True)
        #
        # return np.array([secondary_axis, tertiary_axis, primary_axis]).T

    @property
    def axes(self):
        if self._axes is None:
            self._axes = self.get_axes()
        return self._axes

    @property
    def base_axes(self):
        if self._base_axes is None:
            self._base_axes = self.get_axes(base_axes=True)
        return self._base_axes

    def align(self, axes):
        tf = axes @ self.axes.T
        return self.modify(
            axes=axes,
            elements=[e.transform(tf) for e in self.elements]
        )
    def transform(self, tf):
        return self.modify(
            axes=tf @ self.axes,
            elements=[e.transform(tf) for e in self.elements]
        )

    def get_matrices(self, only_class_representatives=True):
        if only_class_representatives:
            mats = self.character_table.matrices
        else:
            mats = self.get_all_character_matrices()
        if self._axes is not None:
            tf = self._axes @ self.base_axes.T
            mats = tf[np.newaxis] @ mats @ tf.T[np.newaxis]
        return mats

    def plot(self,
             figure=None,
             elements=None,
             origin=None,
             inversion_styles=None,
             rotation_styles=None,
             reflection_styles=None,
             improper_rotation_styles=None,
             **opts):
        from McUtils.Plots import Graphics3D

        if figure is None:
            figure = Graphics3D(backend='x3d')

        if elements is None:
            elements = self.elements
        elif dev.str_is(elements, 'all'):
            elements = self.get_symmetry_elements(only_class_representatives=False)
        if inversion_styles is None:
            inversion_styles = {}
        if rotation_styles is None:
            rotation_styles = {}
        if reflection_styles is None:
            reflection_styles = {}
        if improper_rotation_styles is None:
            improper_rotation_styles = rotation_styles
        for e in elements:
            base_styles = (
                inversion_styles
                    if isinstance(e, InversionElement) else
                reflection_styles
                    if isinstance(e, ReflectionElement) else
                rotation_styles
                    if isinstance(e, RotationElement) else
                improper_rotation_styles
                    if isinstance(e, ImproperRotationElement) else
                {}
            )
            if base_styles is False: continue
            styles = dict(
                opts,
                **base_styles
            )
            e.plot(figure, origin=origin, **styles)

        return figure
class NamedPointGroup(PointGroup):
    def __init__(self, name:'str|NamedPointGroups', character_table=None, elements=None, axes=None):
        super().__init__(character_table=character_table, elements=elements, axes=axes)
        self.group = NamedPointGroups(name)
    def modify(self,
               character_table=dev.default,
               elements=dev.default,
               axes=dev.default
               ):
        return type(self)(
            self.group,
            **self.get_modification_kwargs(
            character_table=character_table,
            elements=elements,
            axes=axes
        ))
    def get_name(self):
        return self.group.value

    def get_character_table(self):
        return CharacterTable.fixed_size_point_group(self.group.value)
    def get_all_character_matrices(self):
        return point_group_data(self.group.value, prop="matrices")

class ParametrizedPointGroup(PointGroup):
    group: ParametrizedPointGroups
    n: int
    def __init__(self, name:'str|ParametrizedPointGroups', n, character_table=None, elements=None, axes=None):
        super().__init__(character_table=character_table, elements=elements, axes=axes)
        self.group = ParametrizedPointGroups(name)
        self.n = n
    def modify(self,
               character_table=dev.default,
               elements=dev.default,
               axes=dev.default
               ):
        return type(self)(
            self.group,
            self.n,
            **self.get_modification_kwargs(
                character_table=character_table,
                elements=elements,
                axes=axes
            ))
    def get_name(self):
        return self.group.value[:1] + str(self.n) + self.group.value[1:]

    def get_character_table(self):
        return CharacterTable.point_group(self.group.value, self.n)
    def get_all_character_matrices(self):
        return point_group_data(self.group.value, self.n, prop="matrices")


